"""
Created on 01 March 2016
@author: Mojtaba Haghighatlari
"""

import pandas as pd
import warnings
import os
from ..utils import std_datetime_str, bool_formatter


class Dragon(object):
    """
    An interface to Dragon 6 and 7 chemoinformatics software.

    Parameters
    ----------
    version: int, optional (default=7)
        The version of available Dragon on the user's system. (available versions: 6 or 7)

    Weights: list, optional (default=["Mass","VdWVolume","Electronegativity","Polarizability","Ionization","I-State"])
        A list of weights to be used

    blocks: list, optional (default=False)
        A list of descriptor blocks' id. For all of them parameter SelectAll="true" is given.
        To select descriptors one by one based on descriptor names, use Script Wizard in Drgon GUI.

    external: boolean, optional (default=False)
        If True, include external variables at the end of each saved file.

    Notes
    -----
        The documentation for the rest of parameters can be found in the following links:
            - http://www.talete.mi.it/help/dragon_help/index.html
            - https://chm.kode-solutions.net/products_dragon_tutorial.php

        In the current version, we recommend the user to use this class with the following parameters:
        molInput = 'file'
        molfile = path to one SMILES representation file with .smi format or a dictionary of many filepaths
        script = 'new'

    Examples
    --------
        >>> import pandas as pd
        >>> from cheml.chem import Dragon
        >>> drg = Dragon(**parameters)
        >>> drg.script_wizard(script='new', output_directory='./')
        >>> drg.run()
        >>> df_path = drg.data_path  # path to the output file
        >>> df = pd.read_csv(df_path, sep=None, engine='python')
        >>> df = df.drop(['No.','NAME'],axis=1)
    """
    def __init__(self, version = 6,CheckUpdates = True,SaveLayout = True, PreserveTemporaryProjects = True,
                ShowWorksheet = False,Decimal_Separator = ".",Missing_String = "NaN",
                DefaultMolFormat = "1",HelpBrowser = "/usr/bin/xdg-open",RejectUnusualValence = False,
                Add2DHydrogens = False,MaxSRforAllCircuit = "19",MaxSR = "35",
                MaxSRDetour = "30",MaxAtomWalkPath = "2000",LogPathWalk = True,
                LogEdge = True,Weights = ["Mass","VdWVolume","Electronegativity","Polarizability","Ionization","I-State"],
                SaveOnlyData = False,SaveLabelsOnSeparateFile = False,SaveFormatBlock = "%b-%n.txt",
                SaveFormatSubBlock = "%b-%s-%n-%m.txt",SaveExcludeMisVal = False,SaveExcludeAllMisVal = False,
                SaveExcludeConst = False,SaveExcludeNearConst = False,SaveExcludeStdDev = False,
                SaveStdDevThreshold = "0.0001",SaveExcludeCorrelated = False,SaveCorrThreshold = "0.95",
                SaveExclusionOptionsToVariables = False,SaveExcludeMisMolecules = False,
                SaveExcludeRejectedMolecules = False,blocks = range(1,30),molInput = "file",
                molInputFormat = "SMILES",molFile = 'required_required',SaveStdOut = False,SaveProject = False,
                SaveProjectFile = "Dragon_project.drp",SaveFile = True,SaveType = "singlefile",
                SaveFilePath = "Dragon_descriptors.txt",logMode = "file",logFile = "Dragon_log.txt",
                external = False,fileName = None,delimiter = ",",consecutiveDelimiter = False,MissingValue = "NaN",
                RejectDisconnectedStrucuture = False, RetainBiggestFragment = False, DisconnectedCalculationOption = "0",
                RoundCoordinates = True, RoundWeights = True, RoundDescriptorValues = True, knimemode = False):
        self.version = version
        self.CheckUpdates = CheckUpdates
        self.PreserveTemporaryProjects = PreserveTemporaryProjects
        self.SaveLayout = SaveLayout
        self.ShowWorksheet = ShowWorksheet
        self.Decimal_Separator = Decimal_Separator
        self.Missing_String = Missing_String
        self.DefaultMolFormat = DefaultMolFormat
        self.HelpBrowser = HelpBrowser
        self.RejectUnusualValence = RejectUnusualValence
        self.Add2DHydrogens = Add2DHydrogens
        self.MaxSRforAllCircuit = MaxSRforAllCircuit
        self.MaxSR = MaxSR
        self.MaxSRDetour = MaxSRDetour
        self.MaxAtomWalkPath = MaxAtomWalkPath
        self.LogPathWalk = LogPathWalk
        self.LogEdge = LogEdge
        self.Weights = Weights
        self.SaveOnlyData = SaveOnlyData
        self.SaveLabelsOnSeparateFile = SaveLabelsOnSeparateFile
        self.SaveFormatBlock = SaveFormatBlock
        self.SaveFormatSubBlock = SaveFormatSubBlock
        self.SaveExcludeMisVal = SaveExcludeMisVal
        self.SaveExcludeAllMisVal = SaveExcludeAllMisVal
        self.SaveExcludeConst = SaveExcludeConst
        self.SaveExcludeNearConst = SaveExcludeNearConst
        self.SaveExcludeStdDev = SaveExcludeStdDev
        self.SaveStdDevThreshold = SaveStdDevThreshold
        self.SaveExcludeCorrelated = SaveExcludeCorrelated
        self.SaveCorrThreshold = SaveCorrThreshold
        self.SaveExclusionOptionsToVariables = SaveExclusionOptionsToVariables
        self.SaveExcludeMisMolecules = SaveExcludeMisMolecules
        self.SaveExcludeRejectedMolecules = SaveExcludeRejectedMolecules
        self.blocks = blocks
        self.molInput = molInput
        self.molInputFormat = molInputFormat
        self.molFile = molFile
        self.SaveStdOut = SaveStdOut
        self.SaveProject = SaveProject
        self.SaveProjectFile = SaveProjectFile
        self.SaveFile = SaveFile
        self.SaveType = SaveType
        self.SaveFilePath = SaveFilePath
        self.logMode = logMode
        self.logFile = logFile
        self.external = external
        self.fileName = fileName
        self.delimiter = delimiter
        self.consecutiveDelimiter = consecutiveDelimiter
        self.MissingValue = MissingValue
        self.RejectDisconnectedStrucuture = RejectDisconnectedStrucuture
        self.RetainBiggestFragment = RetainBiggestFragment
        self.DisconnectedCalculationOption = DisconnectedCalculationOption
        self.RoundCoordinates = RoundCoordinates
        self.RoundWeights = RoundWeights
        self.RoundDescriptorValues = RoundDescriptorValues
        self.knimemode = knimemode

    def script_wizard(self, script='new', output_directory='./'):
        """
        The script_wizard is designed to build a Dragon script file. The name and
        the functionality of this function is the same as available Script wizard
        in the Dragon Graphic User Interface.
        Note: All reported nodes are mandatory, except the <EXTERNAL> tag
        Note: Script for version 7 doesn't support fingerprints block

        Parameters
        ----------
        script: string, optional (default="new")
            If "new" start with a new script. If you want to load an existing script,
            pass the filename with drs format.

        output_directory: string, optional (default = './')
            the path to the working directory to store output files.

        dragon: xml element
            Dragon script in  xml format.

        drs: string
            Dragon script file name

        data_path: string
            The path+name of saved data file in any format. If saveType is 'block'
            or 'subblock' data_path is just the path to the directory that all data
            files have been saved.

        Returns
        -------
            class parameters
        """
        from lxml import objectify, etree
        if output_directory[-1] == '/':
            self.output_directory = output_directory
        else:
            self.output_directory = output_directory + '/'
        if not os.path.exists(self.output_directory):
            os.makedirs(self.output_directory)

        if script == 'new':
            if self.version == 6:
                self.dragon = objectify.Element("DRAGON", version="%i.0.0"%self.version,  script_version="1", generation_date=std_datetime_str('date').replace('-','/'))

                OPTIONS = objectify.SubElement(self.dragon, "OPTIONS")
                OPTIONS.append(objectify.Element("CheckUpdates", value = bool_formatter(self.CheckUpdates)))
                OPTIONS.append(objectify.Element("SaveLayout", value = bool_formatter(self.SaveLayout)))
                OPTIONS.append(objectify.Element("ShowWorksheet", value = bool_formatter(self.ShowWorksheet)))
                OPTIONS.append(objectify.Element("Decimal_Separator", value = self.Decimal_Separator))
                OPTIONS.append(objectify.Element("Missing_String", value = self.Missing_String))
                OPTIONS.append(objectify.Element("DefaultMolFormat", value = self.DefaultMolFormat))
                OPTIONS.append(objectify.Element("HelpBrowser", value = self.HelpBrowser))
                OPTIONS.append(objectify.Element("RejectUnusualValence", value = bool_formatter(self.RejectUnusualValence)))
                OPTIONS.append(objectify.Element("Add2DHydrogens", value = bool_formatter(self.Add2DHydrogens)))
                OPTIONS.append(objectify.Element("MaxSRforAllCircuit", value = self.MaxSRforAllCircuit))
                OPTIONS.append(objectify.Element("MaxSR", value = self.MaxSR))
                OPTIONS.append(objectify.Element("MaxSRDetour", value = self.MaxSRDetour))
                OPTIONS.append(objectify.Element("MaxAtomWalkPath", value = self.MaxAtomWalkPath))
                OPTIONS.append(objectify.Element("LogPathWalk", value = bool_formatter(self.LogPathWalk)))
                OPTIONS.append(objectify.Element("LogEdge", value = bool_formatter(self.LogEdge)))
                Weights = objectify.SubElement(OPTIONS, "Weights")
                for weight in self.Weights:
                    if weight not in ["Mass","VdWVolume","Electronegativity","Polarizability","Ionization","I-State"]:
                        msg = "'%s' is not a valid weight type."%weight
                        raise ValueError(msg)
                    Weights.append(objectify.Element('weight', name = weight))
                OPTIONS.append(objectify.Element("SaveOnlyData", value = bool_formatter(self.SaveOnlyData)))
                OPTIONS.append(objectify.Element("SaveLabelsOnSeparateFile", value = bool_formatter(self.SaveLabelsOnSeparateFile)))
                OPTIONS.append(objectify.Element("SaveFormatBlock", value = self.SaveFormatBlock))
                OPTIONS.append(objectify.Element("SaveFormatSubBlock", value = self.SaveFormatSubBlock))
                OPTIONS.append(objectify.Element("SaveExcludeMisVal", value = bool_formatter(self.SaveExcludeMisVal)))
                OPTIONS.append(objectify.Element("SaveExcludeAllMisVal", value = bool_formatter(self.SaveExcludeAllMisVal)))
                OPTIONS.append(objectify.Element("SaveExcludeConst", value = bool_formatter(self.SaveExcludeConst)))
                OPTIONS.append(objectify.Element("SaveExcludeNearConst", value = bool_formatter(self.SaveExcludeNearConst)))
                OPTIONS.append(objectify.Element("SaveExcludeStdDev", value = bool_formatter(self.SaveExcludeStdDev)))
                OPTIONS.append(objectify.Element("SaveStdDevThreshold", value = self.SaveStdDevThreshold))
                OPTIONS.append(objectify.Element("SaveExcludeCorrelated", value = bool_formatter(self.SaveExcludeCorrelated)))
                OPTIONS.append(objectify.Element("SaveCorrThreshold", value = self.SaveCorrThreshold))
                OPTIONS.append(objectify.Element("SaveExclusionOptionsToVariables", value = bool_formatter(self.SaveExclusionOptionsToVariables)))
                OPTIONS.append(objectify.Element("SaveExcludeMisMolecules", value = bool_formatter(self.SaveExcludeMisMolecules)))
                OPTIONS.append(objectify.Element("SaveExcludeRejectedMolecules", value = bool_formatter(self.SaveExcludeRejectedMolecules)))

                DESCRIPTORS = objectify.SubElement(self.dragon, "DESCRIPTORS")
                for i in self.blocks:
                    if i<1 or i>29:
                        msg = "block id must be in range 1 to 29."
                        raise ValueError(msg)
                    DESCRIPTORS.append(objectify.Element('block', id = "%i"%i, SelectAll = "true"))

                MOLFILES = objectify.SubElement(self.dragon, "MOLFILES")
                MOLFILES.append(objectify.Element("molInput", value = self.molInput))
                if self.molInput == "stdin":
                    if self.molInputFormat not in ['SYBYL','MDL','HYPERCHEM','SMILES','MACROMODEL']:
                        msg = "'%s' is not a valid molInputFormat. Formats:['SYBYL','MDL','HYPERCHEM','SMILES','MACROMODEL']"%self.molInputFormat
                        raise ValueError(msg)
                    MOLFILES.append(objectify.Element("molInputFormat", value = self.molInputFormat))
                elif self.molInput == "file":
                    if isinstance(self.molFile,dict):
                        for f in range(1,len(self.molFile)+1):
                            MOLFILES.append(objectify.Element("molFile",value=self.molFile[f]['file']))
                    elif isinstance(self.molFile,str):
                        MOLFILES.append(objectify.Element("molFile", value=self.molFile))
                    else:
                        msg='Variable molInput can be either a string or a list'
                        raise ValueError(msg)
                else:
                    msg = "Enter a valid molInput: 'stdin' or 'file'"
                    raise ValueError(msg)
                OUTPUT = objectify.SubElement(self.dragon, "OUTPUT")
                OUTPUT.append(objectify.Element("SaveStdOut", value = bool_formatter(self.SaveStdOut)))
                OUTPUT.append(objectify.Element("SaveProject", value = bool_formatter(self.SaveProject)))
                if self.SaveProject:
                    OUTPUT.append(objectify.Element("SaveProjectFile", value = self.SaveProjectFile))
                OUTPUT.append(objectify.Element("SaveFile", value = bool_formatter(self.SaveFile)))
                if self.SaveFile:
                    OUTPUT.append(objectify.Element("SaveType", value = self.SaveType)) # value = "[singlefile/block/subblock]"
                    OUTPUT.append(objectify.Element("SaveFilePath", value = self.output_directory+self.SaveFilePath)) #Specifies the file name for saving results as a plan text file(s), if the "singlefile" option is set; if "block" or "subblock" are set, specifies the path in which results files will be saved.
                OUTPUT.append(objectify.Element("logMode", value = self.logMode)) # value = [none/stderr/file]
                if self.logMode == "file":
                    OUTPUT.append(objectify.Element("logFile", value = self.output_directory+self.logFile))

                if self.external:
                    EXTERNAL = objectify.SubElement(self.dragon, "EXTERNAL")
                    EXTERNAL.append(objectify.Element("fileName", value = self.fileName))
                    EXTERNAL.append(objectify.Element("delimiter", value = self.delimiter))
                    EXTERNAL.append(objectify.Element("consecutiveDelimiter", value = bool_formatter(self.consecutiveDelimiter)))
                    EXTERNAL.append(objectify.Element("MissingValue", value = self.MissingValue))
                self._save_script()

            elif self.version == 7:
                self.dragon = objectify.Element("DRAGON", version="%i.0.0"%self.version, description="Dragon7 - FP1 - MD5270", script_version="1", generation_date=std_datetime_str('date').replace('-','/'))

                OPTIONS = objectify.SubElement(self.dragon, "OPTIONS")
                OPTIONS.append(objectify.Element("CheckUpdates", value = bool_formatter(self.CheckUpdates)))
                OPTIONS.append(objectify.Element("PreserveTemporaryProjects", value = bool_formatter(self.PreserveTemporaryProjects)))
                OPTIONS.append(objectify.Element("SaveLayout", value = bool_formatter(self.SaveLayout)))
#                 OPTIONS.append(objectify.Element("ShowWorksheet", value = bool_formatter(self.ShowWorksheet)))
                OPTIONS.append(objectify.Element("Decimal_Separator", value = self.Decimal_Separator))
                if self.Missing_String == "NaN": self.Missing_String = "na"
                OPTIONS.append(objectify.Element("Missing_String", value = self.Missing_String))
                OPTIONS.append(objectify.Element("DefaultMolFormat", value = self.DefaultMolFormat))
#                 OPTIONS.append(objectify.Element("HelpBrowser", value = self.HelpBrowser))
                OPTIONS.append(objectify.Element("RejectDisconnectedStrucuture", value = bool_formatter(self.RejectDisconnectedStrucuture)))
                OPTIONS.append(objectify.Element("RetainBiggestFragment", value = bool_formatter(self.RetainBiggestFragment)))
                OPTIONS.append(objectify.Element("RejectUnusualValence", value = bool_formatter(self.RejectUnusualValence)))
                OPTIONS.append(objectify.Element("Add2DHydrogens", value = bool_formatter(self.Add2DHydrogens)))
                OPTIONS.append(objectify.Element("DisconnectedCalculationOption", value = self.DisconnectedCalculationOption))
                OPTIONS.append(objectify.Element("MaxSRforAllCircuit", value = self.MaxSRforAllCircuit))
#                 OPTIONS.appendm(objectify.Element("MaxSR", value = self.MaxSR))
                OPTIONS.append(objectify.Element("MaxSRDetour", value = self.MaxSRDetour))
                OPTIONS.append(objectify.Element("MaxAtomWalkPath", value = self.MaxAtomWalkPath))
                OPTIONS.append(objectify.Element("RoundCoordinates", value = bool_formatter(self.RoundCoordinates)))
                OPTIONS.append(objectify.Element("RoundWeights", value = bool_formatter(self.RoundWeights)))
                OPTIONS.append(objectify.Element("LogPathWalk", value = bool_formatter(self.LogPathWalk)))
                OPTIONS.append(objectify.Element("LogEdge", value = bool_formatter(self.LogEdge)))
                Weights = objectify.SubElement(OPTIONS, "Weights")
                for weight in self.Weights:
                    if weight not in ["Mass","VdWVolume","Electronegativity","Polarizability","Ionization","I-State"]:
                        msg = "'%s' is not a valid weight type."%weight
                        raise ValueError(msg)
                    Weights.append(objectify.Element('weight', name = weight))
                OPTIONS.append(objectify.Element("SaveOnlyData", value = bool_formatter(self.SaveOnlyData)))
                OPTIONS.append(objectify.Element("SaveLabelsOnSeparateFile", value = bool_formatter(self.SaveLabelsOnSeparateFile)))
                OPTIONS.append(objectify.Element("SaveFormatBlock", value = self.SaveFormatBlock))
                OPTIONS.append(objectify.Element("SaveFormatSubBlock", value = self.SaveFormatSubBlock))
                OPTIONS.append(objectify.Element("SaveExcludeMisVal", value = bool_formatter(self.SaveExcludeMisVal)))
                OPTIONS.append(objectify.Element("SaveExcludeAllMisVal", value = bool_formatter(self.SaveExcludeAllMisVal)))
                OPTIONS.append(objectify.Element("SaveExcludeConst", value = bool_formatter(self.SaveExcludeConst)))
                OPTIONS.append(objectify.Element("SaveExcludeNearConst", value = bool_formatter(self.SaveExcludeNearConst)))
                OPTIONS.append(objectify.Element("SaveExcludeStdDev", value = bool_formatter(self.SaveExcludeStdDev)))
                OPTIONS.append(objectify.Element("SaveStdDevThreshold", value = self.SaveStdDevThreshold))
                OPTIONS.append(objectify.Element("SaveExcludeCorrelated", value = bool_formatter(self.SaveExcludeCorrelated)))
                OPTIONS.append(objectify.Element("SaveCorrThreshold", value = self.SaveCorrThreshold))
                OPTIONS.append(objectify.Element("SaveExclusionOptionsToVariables", value = bool_formatter(self.SaveExclusionOptionsToVariables)))
                OPTIONS.append(objectify.Element("SaveExcludeMisMolecules", value = bool_formatter(self.SaveExcludeMisMolecules)))
                OPTIONS.append(objectify.Element("SaveExcludeRejectedMolecules", value = bool_formatter(self.SaveExcludeRejectedMolecules)))
                OPTIONS.append(objectify.Element("RoundDescriptorValues", value = bool_formatter(self.RoundDescriptorValues)))

                DESCRIPTORS = objectify.SubElement(self.dragon, "DESCRIPTORS")
                for i in self.blocks:
                    if i<1 or i>30:
                        msg = "block id must be in range 1 to 30."
                        raise ValueError(msg)
                    DESCRIPTORS.append(objectify.Element('block', id = "%i"%i, SelectAll = "true"))

                MOLFILES = objectify.SubElement(self.dragon, "MOLFILES")
                MOLFILES.append(objectify.Element("molInput", value = self.molInput))
                if self.molInput == "stdin":
                    if self.molInputFormat not in ['SYBYL','MDL','HYPERCHEM','SMILES','CML','MACROMODEL']:
                        msg = "'%s' is not a valid molInputFormat. Formats:['SYBYL','MDL','HYPERCHEM','SMILES','CML','MACROMODEL']"%self.molInputFormat
                        raise ValueError(msg)
                    MOLFILES.append(objectify.Element("molInputFormat", value = self.molInputFormat))
                elif self.molInput == "file":
                    if isinstance(self.molFile,dict):
                        for f in range(1,len(self.molFile)+1):
                            MOLFILES.append(objectify.Element("molFile",value=self.molFile[f]['file']))
                    elif isinstance(self.molFile,str):
                        MOLFILES.append(objectify.Element("molFile", value=self.molFile))
                    else:
                        msg='Variable molInput can be either a string or a list'
                        raise ValueError(msg)
                else:
                    msg = "Enter a valid molInput: 'stdin' or 'file'"
                    raise ValueError(msg)
                OUTPUT = objectify.SubElement(self.dragon, "OUTPUT")
                OUTPUT.append(objectify.Element("knimemode", value = bool_formatter(self.knimemode)))
                OUTPUT.append(objectify.Element("SaveStdOut", value = bool_formatter(self.SaveStdOut)))
                OUTPUT.append(objectify.Element("SaveProject", value = bool_formatter(self.SaveProject)))
                if self.SaveProject:
                    OUTPUT.append(objectify.Element("SaveProjectFile", value = self.SaveProjectFile))
                OUTPUT.append(objectify.Element("SaveFile", value = bool_formatter(self.SaveFile)))
                if self.SaveFile:
                    OUTPUT.append(objectify.Element("SaveType", value = self.SaveType)) # value = "[singlefile/block/subblock]"
                    OUTPUT.append(objectify.Element("SaveFilePath", value = self.output_directory+self.SaveFilePath)) #Specifies the file name for saving results as a plan text file(s), if the "singlefile" option is set; if "block" or "subblock" are set, specifies the path in which results files will be saved.
                OUTPUT.append(objectify.Element("logMode", value = self.logMode)) # value = [none/stderr/file]
                if self.logMode == "file":
                    OUTPUT.append(objectify.Element("logFile", value = self.output_directory+self.logFile))

                if self.external:
                    EXTERNAL = objectify.SubElement(self.dragon, "EXTERNAL")
                    EXTERNAL.append(objectify.Element("fileName", value = self.fileName))
                    EXTERNAL.append(objectify.Element("delimiter", value = self.delimiter))
                    EXTERNAL.append(objectify.Element("consecutiveDelimiter", value = bool_formatter(self.consecutiveDelimiter)))
                    EXTERNAL.append(objectify.Element("MissingValue", value = self.MissingValue))
                self._save_script()

            else:
                msg = "Only version 6 and version 7 (newest version) are available in this module."
                warnings.warn(msg,Warning)

        else:
            doc = etree.parse(script)
            self.dragon = etree.tostring(doc) 	# dragon script : dragon
            self.dragon = objectify.fromstring(self.dragon)
            objectify.deannotate(self.dragon)
            etree.cleanup_namespaces(self.dragon)
            if self.dragon.attrib['version'][0] not in ['6','7']:
                msg = "Dragon script is not labeled to the newest vesions of Dragon, 6 or 7. This may causes some problems."
                warnings.warn(msg,Warning)
            mandatory_nodes = ['OPTIONS','DESCRIPTORS','MOLFILES','OUTPUT']
            reported_nodes = [element.tag for element in self.dragon.iterchildren()]
            if not set(reported_nodes).issuperset(set(mandatory_nodes)):
                msg = 'Dragon script does not contain all mandatory nodes, which are:%s'%str(mandatory_nodes)
                raise ValueError(msg)
            self.drs = script
        self.data_path = self.dragon.OUTPUT.SaveFilePath.attrib['value']
        return self

    def _save_script(self):
        from lxml import objectify, etree
        objectify.deannotate(self.dragon)
        etree.cleanup_namespaces(self.dragon)
        self.drs_name = 'Dragon_script.drs'
        with open(self.output_directory+self.drs_name, 'w') as outfile:
            outfile.write("%s" %etree.tostring(self.dragon, pretty_print=True))

    def printout(self):
        from lxml import objectify, etree
        objectify.deannotate(self.dragon)
        etree.cleanup_namespaces(self.dragon)
        print(objectify.dump(self.dragon))

    def run(self):
        print "running Dragon%s ..."%self.version
        os.system('nohup dragon%sshell -s %s'%(self.version,self.output_directory+self.drs_name))
        # print subprocess.check_output(['nohup dragon%sshell -s %s'%(self.version,self.drs)])
        print "... Dragon job completed!"



import types
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class ElementFractionAttributeGenerator:
    """Class to set the element fractions as the features of materials.
    """

    def generate_features(self, entries):
        """
        Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert feature headers here.
        for elem in LookUpData.element_names:
            feat_headers.append("X_"+elem)

        # Insert feature values here.
        for entry in entries:
            tmp_list = []
            elem_ids = entry.get_element_ids()
            elem_fracs = entry.get_element_fractions()
            for elem in range(len(LookUpData.element_names)):
                if elem in elem_ids:
                    idx = elem_ids.index(elem)
                    tmp_list.append(elem_fracs[idx])
                else:
                    tmp_list.append(0.0)
            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import numpy as np
import pandas as pd
import types
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry

class ChemicalOrderingAttributeGenerator:
    """Class to compute attributes based on chemical ordering of structure.

    Determines average Warren-Cowley ordering parameter for the bond network
    defined by the Voronoi tessellation of a structure.

    Attributes
    ----------
    shells : list
        Index of shells to compute features for.
    weighted : bool
        Whether to compute features using weighting or not.

    See Also
    --------
    VoronoiCellBasedAnalysis.get_neighbor_ordering_parameters : Computes
    Warren-Cowley ordering parameters.

    Notes
    -----
    For each atom in the structure, the average Warren-Cowley ordering
    parameter is determined by computing the average magnitude of ordering
    parameter for each type for all atoms in a structure. The ordering
    parameter is 0 for a perfectly-random distribution, so this average
    represents an average degree of "ordering" in the structure. This
    attribute is computed for several nearest-neighbor shells (1st, 2nd,
    and 3rd by default).

    There are two options for computing order parameters: Weighted and
    unweighted. The former is computed by weighing the contribution of each
    neighboring atom by the fraction of surface area corresponding to
    boundaries between that atom and the central atom. The former considers
    all neighbors weighted equally, which means they are very sensitive to
    the introduction of small faces due to numerical problems inherent to the
    Voronoi tessellation. Full details is available in the Vassal
    documentation for VoronoiCellBasedAnalysis.getNeighborOrderingParameters().

    """

    def __init__(self):
        """Function to create instance and initialize fields.

        Will create the WC parameters for the first, second and third
        nearest-neighbor shells by default.

        """

        # Shells to compute the WC attribute for.
        self.shells = [1, 2, 3]

        # Whether to compute weighted WC ordering parameters.
        self.weighted = True

    def set_shells(self, shells):
        """Function to set which nearest-neighbor shells to consider when
        generating features.

        Parameters
        ----------
        shells: list
            Desired shell indices.

        """

        self.shells = list(shells)

    def set_weighted(self, weighted):
        """Function to set whether to consider face sizes when computing
        ordering parameters.

        Parameters
        ----------
        weighted: bool
            Whether to weigh using face sizes.

        """

        self.weighted = weighted

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize list of feature values for pandas data frame.
        feat_values = []

        # Raise exception if input argument is not of type list of
        # CrystalStructureEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Insert header names here.
        feat_headers = ["mean_WCMagnitude"+ ("" if self.weighted else
                        "_unweighted_") + "_shell" + str(shell) for shell in
                        self.shells]

        # Compute features.
        for entry in entries:
            tmp_list = []
            # Get the Voronoi tessellation.
            try:
                voro = entry.compute_voronoi_tessellation()
            except Exception:
                tmp_list = [np.nan] * len(feat_headers)
                feat_values.append(tmp_list)
                continue

            tmp_list = [voro.warren_cowley_ordering_magnitude(shell=shell,
                        weighted=self.weighted) for shell in self.shells]
            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


class Input(object):
    """Container class for inputs of a function"""
    def __init__(self,name, short_description, types):
        self.name = name
        self.short_description = short_description
        self.long_description = ''
        self.value = None
        self.types = types

class Output(object):
    """Container class for outputs of a function"""
    def __init__(self,name, short_description, types):
        self.name = name
        self.short_description = short_description
        self.long_description = ''
        self.value = None
        self.types = types
        self.fro = ('iblock', 'host', 'function')
        self.count = 0


class Parameter(object):
    """Container class for parameters of a function"""
    def __init__(self,name,default,format='',required=False,description="",options=[]):
        self.name = name
        self.default = default
        self.format = format
        self.required = required
        self.description = description
        self.options = options

def req(ind):
    all_req = { 0 : ('ChemML','0.4.1'),
                1 : ('scikit-learn','0.19.0'),
                2 : ('pandas', '0.20.3'),
                3 : ('RDKit','2016.03.1'),
                4 : ('Dragon','7 or 6'),
                5 : ('lxml', '3.4.0'),
                6 : ('Babel', '2.3.4'),
                7 : ('matplotlib','1.5.1'),
                8 : ('keras', '2.1.2'),
                9 : ('tensorflow', '1.4.1'),
                10: ('deap', '1.2.2')
              }
    return all_req[ind]

#conda install -c conda-forge -c rdkit numpy pandas scikit-learn tensorflow keras rdkit babel deap matplotlib lxml ipywidgets graphviz


def regression_types():
    """all the regression classes that follow the sklearn format"""
    sklearn_types = ["<class 'sklearn.linear_model.base.LinearRegression'>","<class 'sklearn.linear_model.ridge.Ridge'>",
                 "<class 'sklearn.kernel_ridge.KernelRidge'>", "<class 'sklearn.linear_model.coordinate_descent.Lasso'>",
                 "<class 'sklearn.linear_model.coordinate_descent.MultiTaskLasso'>", "<class 'sklearn.linear_model.coordinate_descent.ElasticNet'>",
                 "<class 'sklearn.linear_model.coordinate_descent.MultiTaskElasticNet'>", "<class 'sklearn.linear_model.least_angle.Lars'>",
                 "<class 'sklearn.linear_model.least_angle.LassoLars'>", "<class 'sklearn.linear_model.bayes.BayesianRidge'>",
                 "<class 'sklearn.linear_model.bayes.ARDRegression'>", "<class 'sklearn.linear_model.logistic.LogisticRegression'>",
                 "<class 'sklearn.linear_model.stochastic_gradient.SGDRegressor'>", "<class 'sklearn.svm.classes.SVR'>",
                 "<class 'sklearn.svm.classes.NuSVR'>", "<class 'sklearn.svm.classes.LinearSVR'>", "<class 'sklearn.neural_network.multilayer_perceptron.MLPRegressor'>"]
    cheml_types = ["<class 'cheml.nn.keras.mlp.MLP_sklearn'>"]
    all_types = tuple(sklearn_types+cheml_types)
    return all_types

def cv_classes():
    """all objects to be used as a cross-validation generator"""
    all_types = ("<class 'sklearn.model_selection._split.KFold'>", "<class 'sklearn.model_selection._split.ShuffleSplit'>",
                    "<class 'sklearn.model_selection._split.StratifiedShuffleSplit'>",
                 "<class 'sklearn.model_selection._split.LeaveOneOut'>")

    return all_types

###########

def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result




import numpy as np
from numpy.linalg import norm
from ..data.AtomImage import AtomImage
from ..geometry.Plane import Plane
from ..util.VectorCombinationComputer import VectorCombinationComputer

class PairDistanceAnalysis:
    """Class to compute the distances between each pairs of atom. Determines the
    distance between each of atoms (and any periodic image within a certain
    distance).

    Attributes
    ----------
    structure : Cell
        Link to structure being evaluated.
    cut_off_distance : float
        Cutoff distance. No pairs farther than this value are considered.
    supercells : list
        Periodic images to consider when searching for neighbors.
    lattice_vectors : list
        Lattice vectors corresponding to each image.

    """

    def __init__(self):
        """Function to instantiate the class.

        Parameters
        ----------
        structure : Cell
            Link to structure being evaluated.
        cut_off_distance : float
            Cutoff distance. No pairs farther than this value are considered.
        supercells : list
            Periodic images to consider when searching for neighbors.
        lattice_vectors : list
            Lattice vectors corresponding to each image.
        """

        # Cutoff distance. No pairs farther than this value are considered.
        self.cutoff_distance = 6.5

        # Periodic images to consider when searching for neighbors.
        self.supercells = []

        # Lattice vectors corresponding to each image.
        self.lattice_vectors = []

        # Link to structure being evaluated.
        self.structure = None

    def precompute(self):
        """Function to perform any kind of computations that should be
        performed only once.

        """

        lat_vectors = self.structure.get_lattice_vectors()
        # Goal: Compute all combinations of lattice vectors that are less
        # than the cutoff length plus the length of the longest vector.
        #
        # Find the largest distance between an atom and one of the vertices
        # of the unit cell. Computed using the Voronoi tessellation. This
        # point is the vertex corresponding to the faces of (origin, {a & b &
        #  c}).
        p0 = Plane(normal=lat_vectors[0], tolerance=1e-6, p= 0.5 * lat_vectors[0])
        p1 = Plane(normal=lat_vectors[1], tolerance=1e-6, p= 0.5 * lat_vectors[1])
        p2 = Plane(normal=lat_vectors[2], tolerance=1e-6, p= 0.5 * lat_vectors[2])
        x = Plane.intersection_3_planes(p0, p1, p2)
        max_image_dist = norm(x)

        # In order to create a list of vectors such the set of displacement
        # vectors from any atom in the structure to every image of another
        # atom is within the cutoff radius is a subset of the set of
        # displacement vectors created by adding the displacement vector of
        # the closest image of the second atom from the first atom to each
        # atom in the list of vectors, we need to find all vectors whose
        # length is shorter than max_dist + cutoff_distance.
        #
        # Rationale:
        # 1. The maximum distance between an atom and the closet neighbor
        # is no more than maxDist (defined above). Otherwise, the image is
        # not the closest image.
        # 2. If we are interested in finding all images of that second atom
        # that are within the cutoff distance of the first atom, the distance
        # between any of those images and the closest image of the second
        # atom can't be greater than the cutoff distance plus the distance
        # between that closest image and the first atom.
        # 3. Those images of the second atoms are located at integer
        # multiples of the lattice vectors. So, if we find all lattice
        # vectors shorter than cutoff_distance + longest possible distance
        # between an atom and the closest image of a second atom, we can find
        #  all images!

        # Compute those vectors.
        computer = VectorCombinationComputer(lat_vectors, max_image_dist +
                                             self.cutoff_distance)
        self.supercells = computer.get_supercell_coordinates()
        self.lattice_vectors = computer.get_vectors()

    def get_cutoff_distance(self):
        """Function to get the cutoff distance.

        Returns
        -------
        cutoff_distance : float
            Cutoff distance.

        """
        return self.cutoff_distance

    def set_cutoff_distance(self, d):
        """Function to set the cutoff distance.

        Parameters
        ----------
        d : float
            Desired cutoff distance.

        """
        self.cutoff_distance = d
        if self.structure is not None:
            self.precompute()

    def find_all_images(self, center_atom, neighbor_atom):
        """Function to find all images of one atom that are within the cutoff
        distance of another atom.

        Parameters
        ----------
        center_atom : int
            Index of atom at the center.
        neighbor_atom : int
            Neighboring atom index.

        Returns
        -------
        output : list
            All images of neighbor_atom that within a cutoff distance of
            center_atom, and their respective distances. List of tuples (
            neighboring atom, distance).

        """

        # Get the two atoms.
        center_pos = self.structure.get_atom(
            center_atom).get_position_cartesian()

        # Find all images.
        output = []
        closest_image = self.structure.get_minimum_distance(
            center=center_atom, neighbor=neighbor_atom)

        B_sub_A = closest_image.get_position()
        B_sub_A -= center_pos

        closest_supercell = closest_image.get_supercell()

        # Loop over each periodic image to find those within range.
        cutoff_distance_sq = self.cutoff_distance ** 2
        for img in range(len(self.supercells)):
            supercell_vec = self.lattice_vectors[img]
            new_pos = B_sub_A.copy() + supercell_vec
            dist = new_pos[0] ** 2 + new_pos[1] ** 2 + new_pos[2] ** 2

            if dist < cutoff_distance_sq and dist > 1e-8:
                ss = self.supercells[img].copy() + closest_supercell
                d = np.math.sqrt(dist)
                pos = closest_image.get_atom().get_position()
                output.append((AtomImage(closest_image.get_atom(), ss), d))

        return output

    def get_all_neighbors_of_atom(self, index):
        """Function to compute all neighbors of a certain atom.

        Parameters
        ----------
        index : int
            Index of atom being considered.

        Returns
        -------
        output : list
            List of tuples (neighboring atom, distance).

        """
        output = []
        for atom in self.structure.get_atoms():
            output += self.find_all_images(index, atom.get_id())
        return output

    def compute_PRDF(self, n_bin):
        """Function to compute the pair radial distribution function.

        Parameters
        ----------
        n_bin : int
            Number of bins in which to discretize PRDF (must be > 0)

        Returns
        -------
        output : array-like
            Pair radial distribution function between each type. Bin i,
            j,k is the density of bonds between types i and j that are between k
            * cutoff / nBin and (k + 1) * cutoff / nBin.

        Raises
        ------
        ValueError
            If n_bin is less than or equal to 0.
        """
        if n_bin <= 0:
            raise ValueError("Number of bins must be 0.")

        n_t = self.structure.n_types()
        n_a = self.structure.n_atoms()

        # Initialize arrays.
        output = np.zeros((n_t, n_t, n_bin))
        # Find all pairs within the cutoff distance.
        n_type = np.zeros(n_t, dtype=int)

        for i in range(n_a):
            for j in range(i, n_a):
                # Find images.
                images = self.find_all_images(i, j)
                i_type = self.structure.get_atom(i).get_type()
                n_type[i_type] += 1
                j_type = self.structure.get_atom(j).get_type()

                # For each image, assign it to bin.
                for img in images:
                    bin_ = int(np.math.floor(img[1] * n_bin /
                                             self.cutoff_distance))
                    if bin_ >= n_bin:
                        # Happens if dist equals cutoff.
                        continue
                    output[i_type][j_type][bin_] += 1
                    output[j_type][i_type][bin_] += 1

        # Normalizing data.
        bin_spacing = self.cutoff_distance / n_bin
        for b in range(n_bin):
            vol = 4.0 / 3.0 * ((b + 1) ** 3 - b ** 3) * bin_spacing ** 3 * \
                  np.math.pi
            for i in range(n_t):
                for j in range(n_t):
                    output[i][j][b] /= vol * n_type[i]

        return output

    def analyze_structure(self, s):
        """Function to analyze a specific structure. Once this completes,
        it is possible to retrieve results out of this object.

        Parameters
        ----------
        s : Cell
            Structure to be analyzed.

        """
        self.structure = s
        self.precompute()

    def recompute(self):
        """Function to recompute structural information.

        """
        self.precompute()


import pandas as pd
import numpy as np
import os
import fnmatch

# TODO: AP/bit is time consuming - should be used directly with distance matrix

def _add_range(file,start_end):
    start = start_end[0::2]
    end = start_end[1::2]
    nit = file.count('%s')
    list_ranges = [str(range(start[it],end[it]+1))[1:-1] for it in range(len(start))]
    s = "pattern = '"+file+"'" + '%(' + str(list_ranges)[1:-1] + ')'
    exec(s)
    return pattern

class RDKitFingerprint(object):
    """ An interface to RDKit fingerprints.
    
    Parameters
    ----------
    FPtype: string, optional (default='Morgan')
        Available fingerprint types:
            - 'Hashed_atom_pair' or 'HAP' 
            - 'Atom_pair' or 'AP'
            - 'MACCS'
            - 'Morgan'
            - 'Hashed_topological_torsion' or 'HTT'
            - 'Topological_torsion' or 'TT' 

    vector: string, optional (default = 'bit')
        Available options for vector:
            - 'int' : including counts for each bit instead of just zeros and ones
                    It is not available for 'MACCS'.
            - 'bit' : only zeros and ones
                    It is not available for 'Topological_torsion'.

    nBits: integer, optional (default = 1024)
        It sets number of bits in the specific type of fingerprint vectors. 
        Not availble for: 
            - 'MACCS' (MACCS keys have a fixed length of 167 bits)
            - 'Atom_pair' (In the Atom_pair we return only those bits that are 
                           non zero for at least one of the molecules. Thus, the size 
                           will be significantly less compare to more than 8 million  
                           available bits)
            - 'Topological_torsion'  

    radius: integer, optional (default = 2)
        only availble for 'Morgan' fingerprint.

    removeHs: boolean, optional (default=True)
        If True, remove any hydrogen from the graph of a molecule.

    Returns
    -------
    data set of fingerprint vectors.
    """
    def __init__(self, removeHs=True, FPtype='Morgan', vector='bit', nBits=1024, 
                 radius = 2):
        self.FPtype = FPtype
        self.vector = vector
        self.nBits = nBits
        self.radius = radius
        self.removeHs = removeHs

    def MolfromFile(self, molfile, path=None, *arguments):
        """ (MolfromFile)
        Construct molecules from one or more input files.
        
        Parameters
        ----------
        molfile: string
            This is the place you define molecules file name and path. It can contain 
            any special character in the following list:
            
                *       : matches everything
                ?       : matches any single character
                [seq]   : matches any character in seq
                [!seq]  : matches any character not in seq
                /       : filename seperator

            The pattern matching is implemented by fnmatch - Unix filename pattern matching.
            Note: The pattern must include the extension at the end. A list of available 
            file extensions is provided here: 
            
                '.mol','.mol2'  : molecule files 
                '.pdb'          : protein data bank file
                '.tpl'          : TPL file
                '.smi'          : file of one or more lines of Smiles string
                '.smarts'       : file of one or more lines of Smarts string

        path: string, optional (default = None)            
            If path is None, this function tries to open the file as a single
            file (without any pattern matching). Therefore, none of the above 
            special characters except '/' are helpful when they are not part 
            of the file name or path.    
            If not None, it determines the path that this function walk through
            and look for every file that mathes the pattern in the file. To start
            walking from the current directory, the path value should be '.'
        
        *arguments: integer
            If sequences in the special characters should be a range of integers,
            you can easily pass the start and the end arguments of range as extra 
            arguments. But, notice that:
                1- In this case remember to replace seq with %s. Then, you should 
                   have [%s] or [!%s] in the file pattern.
                2- for each %s you must pass two integers (always an even number 
                    of extra arguments must be passed). Also, the order is important
                    and it should be in the same order of '%s's. 
            
        examples
        --------
            (1)
            file: 'Mydir/1f/1_opt.smi'
            path: None
            >>> sample files to be read: 'Mydir/1f/1_opt.smi' 
            
            (2)
            file: '[1,2,3,4]?/*_opt.mol'
            path: 'Mydir'
            >>> sample files to be read: 'Mydir/1f/1_opt.mol', 'Mydir/2c/2_opt.mol', ... 
            
            (3)
            file: '[!1,2]?/*_opt.pdb'
            path: '.'
            >>> sample files to be read: './3f/3_opt.pdb', 'Mydir/7c/7_opt.pdb', ... 

            (4)
            file: '*['f','c']/*_opt.tpl'
            path: 'Mydir'
            >>> sample files to be read: 'Mydir/1f/1_opt.tpl', 'Mydir/2c/2_opt.tpl', ... 

            (5)
            file: '[%s]?/[!%s]_opt.mol2'
            arguments: 1,4,7,10
            path: 'Mydir/all'
            >>> sample files to be read: 'Mydir/all/1f/1_opt.mol2', 'Mydir/all/2c/2_opt.mol2', ... 
        """
        from rdkit import Chem
        extensions = {'.mol':       Chem.MolFromMolFile,
                      '.mol2':      Chem.MolFromMol2File,
                      '.pdb':       Chem.MolFromPDBFile,
                      '.tpl':       Chem.MolFromTPLFile,
                      '.smi':       Chem.MolFromSmiles,
                      '.smarts':    Chem.MolFromSmarts,
                      # '.inchi':     Chem.MolFromInchi
                      }
        file_name, file_extension = os.path.splitext(molfile)
        if file_extension == '':
            msg = 'file extension not determined'
            raise ValueError(msg)
        elif file_extension not in extensions:
            msg = "file extension '%s' not available"%file_extension
            raise ValueError(msg)
        start_end = [arg for arg in arguments]
        if len(start_end) != file_name.count('%s')*2:
            msg = "pass an even number of integers: 2 for each '%s'"
            raise ValueError(msg)        
        
        if path and '%s' in molfile:
            molfile = _add_range(molfile,start_end)

        self.removed_rows = []
        self.molecules = []
        if path:
            for root, directories, filenames in os.walk(path):
                file_path = [os.path.join(root, filename) for filename in filenames]
                for filename in fnmatch.filter(file_path, os.path.join(path, molfile)):
                    if file_extension in ['.smi','.smarts']:
                        mols = open(filename, 'r')
                        mols = mols.readlines()
                        for i, x in enumerate(mols):
                            mol = extensions[file_extension](x.strip(),removeHs=self.removeHs)
                            if mol is None:
                                self.removed_rows.append(i)
                            else:
                                self.molecules.append(mol)
                        # mols = [extensions[file_extension](x.strip(),removeHs=self.removeHs) for x in mols]
                        # self.molecules += mols
                    else:
                        self.molecules += [extensions[file_extension](filename,removeHs=self.removeHs)]
        else:
            if file_extension in ['.smi','.smarts']:
                mols = open(molfile, 'r')
                mols = mols.readlines()
                for i,x in enumerate(mols):
                    mol = extensions[file_extension](x.strip())
                    if mol is None:
                        self.removed_rows.append(i)
                    else:
                       self.molecules.append(mol)
                # mols = [extensions[file_extension](x.strip()) for x in mols]
                # self.molecules += mols
            else:
                self.molecules += [extensions[file_extension](filename,removeHs=self.removeHs)]

    def Fingerprint(self):
        if self.FPtype == 'Hashed_atom_pair' or self.FPtype == 'HAP':
            if self.vector == 'int':
                from rdkit.Chem.AtomPairs.Pairs import GetHashedAtomPairFingerprint
                self.fps = [GetHashedAtomPairFingerprint(m,nBits=self.nBits) for m in self.molecules]
                dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps]
                data = pd.DataFrame(dict_nonzero,columns=range(self.nBits))
                data.fillna(0, inplace = True)
                return data
            elif self.vector == 'bit':
                from rdkit.Chem.rdMolDescriptors import GetHashedAtomPairFingerprintAsBitVect
                self.fps = [GetHashedAtomPairFingerprintAsBitVect(m,nBits=self.nBits) for m in self.molecules]
                data = np.array(self.fps)
                data = pd.DataFrame(data)
                return data
            else:
                msg = "The argument vector can be 'int' or 'bit'"
                raise ValueError(msg)
        elif self.FPtype == 'Atom_pair' or self.FPtype == 'AP':
            if self.vector == 'int':
                from rdkit.Chem.AtomPairs.Pairs import GetAtomPairFingerprint
                self.fps = [GetAtomPairFingerprint(m) for m in self.molecules]
                dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps]
                pairScores = []
                for fp in dict_nonzero:
                    pairScores += [key for key in fp]
                data = pd.DataFrame(dict_nonzero,columns=list(set(pairScores)))
                data.fillna(0, inplace = True)
                return data
            elif self.vector == 'bit':
                from rdkit.Chem.AtomPairs.Pairs import GetAtomPairFingerprintAsBitVect
                self.fps = [GetAtomPairFingerprintAsBitVect(m) for m in self.molecules]
                data = np.array(self.fps)
                data = pd.DataFrame(data)

                print len(data.columns)
                d_des = data.describe()
                for i in data.columns:
                    if d_des[i]['mean'] == 0 :
                        data.drop(i,1)
                print len(data.columns)

                dict_nonzero = []
                for fp in self.fps:
                    dict_nonzero.append({i:el for i,el in enumerate(fp) if el!=0})
                pairScores = []
                for fp in dict_nonzero:
                    pairScores += [key for key in fp]
                data = pd.DataFrame(dict_nonzero,columns=list(set(pairScores)))
                data.fillna(0, inplace = True)
                return data
            else:
                msg = "The argument vector can be 'int' or 'bit'"
                raise ValueError(msg)
        elif self.FPtype == 'MACCS':
            if self.vector == 'int':
                msg = "There is no RDKit function to encode int vectors for MACCS keys"
                raise ValueError(msg)
            elif self.vector == 'bit':
                from rdkit.Chem.MACCSkeys import GenMACCSKeys
                self.fps = [GenMACCSKeys(mol) for mol in self.molecules]
                data = np.array(self.fps)
                data = pd.DataFrame(data)
                return data
            else:
                msg = "The vector argument can only be 'int' or 'bit'"
                raise ValueError(msg)
        elif self.FPtype == 'Morgan':
            if self.vector == 'int':
                from rdkit.Chem.rdMolDescriptors import GetMorganFingerprint
                self.fps = [GetMorganFingerprint(mol, self.radius) for mol in self.molecules]
                dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps]
                pairScores = []
                for fp in dict_nonzero:
                    pairScores += list(fp)
                data = pd.DataFrame(dict_nonzero,columns=list(set(pairScores)))
                data.fillna(0, inplace = True)
                return data
            elif self.vector == 'bit':
                from rdkit.Chem.rdMolDescriptors import GetMorganFingerprintAsBitVect
                self.fps = [GetMorganFingerprintAsBitVect(mol, self.radius,nBits=self.nBits) for mol in self.molecules]
                data = np.array(self.fps)
                data = pd.DataFrame(data)
                return data
            else:
                msg = "The argument vector can only be 'int' or 'bit'"
                raise ValueError(msg)
        elif self.FPtype == 'Hashed_topological_torsion' or self.FPtype == 'HTT':
            if self.vector == 'int':
                from rdkit.Chem.rdMolDescriptors import GetHashedTopologicalTorsionFingerprint
                self.fps = [GetHashedTopologicalTorsionFingerprint(m,nBits=self.nBits) for m in self.molecules]
                dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps]
                data = pd.DataFrame(dict_nonzero,columns=range(self.nBits))
                data.fillna(0, inplace = True)
                return data
            elif self.vector == 'bit':
                from rdkit.Chem.rdMolDescriptors import GetHashedTopologicalTorsionFingerprintAsBitVect
                self.fps = [GetHashedTopologicalTorsionFingerprintAsBitVect(m,nBits=self.nBits) for m in self.molecules]
                data = np.array(self.fps)
                data = pd.DataFrame(data)
                return data
            else:
                msg = "The argument vector can be 'int' or 'bit'"
                raise ValueError(msg)
        elif self.FPtype == 'Topological_torsion' or self.FPtype == 'TT':
            if self.vector == 'int':
                from rdkit.Chem.AtomPairs.Torsions import GetTopologicalTorsionFingerprintAsIntVect
                self.fps = [GetTopologicalTorsionFingerprintAsIntVect(mol) for mol in self.molecules]
                dict_nonzero = [fp.GetNonzeroElements() for fp in self.fps]
                pairScores = []
                for fp in dict_nonzero:
                    pairScores += list(fp)
                data = pd.DataFrame(dict_nonzero,columns=list(set(pairScores)))
                data.fillna(0, inplace = True)
                return data
            elif self.vector == 'bit':
                msg = "There is no RDKit function to encode bit vectors for Topological Torsion Fingerprints"
                raise ValueError(msg)
            else:
                msg = "The argument vector can only be 'int'"
                raise ValueError(msg)
        else:
            msg = "The type argument '%s' is not a valid fingerprint type"%self.FPtype
            raise ValueError(msg)
    
    # def Similarity(self,data=None, metric='Euclidean', nCores = 1):
    #     """(Similarity)
    #     Calculate molecular similarity matrix, while each row of data is a feature
    #     vector of one molecule.
    #
    #     Parameters
    #     ----------
    #     data: pandas dataframe, optional (default = None)
    #         Each row of data represents corresponding features of one molecule (sample).
    #
    #     metric: string, optional (default = 'Euclidean')
    #         Similarity metric. Available similarity metrics include Euclidean,
    #         Tanimoto, Dice, Cosine, Sokal, Russel, Kulczynski, McConnaughey, and Tversky.
    #
    #         Note:
    #             if data = None, all metrics except 'Euclidean' can be used. The
    #             Fingerprint function must be called first. Thus, the similarity metrics
    #             implemetned by RDKit will be used on calculated fingerprints.
    #
    #             if a dataframe has been passed for data, the only available metric
    #             is 'Euclidean'.
    #
    #     nCores: integer, optional (default = 1)
    #         Number of cores for multiprocessing.
    #     """
    #     if data == None:
    #         if metric in ['Tanimoto','Dice','Cosine','Sokal','Russel','Kulczynski','McConnaughey','Tversky']:
    #             from rdkit import DataStructs
    #
    #         else:
    #             msg = "The metric '%s' is not a valid metric type for "%metric
    #             raise ValueError(msg)
    #     return None


import types
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class MeredigAttributeGenerator:
    """Class to generate attributes as described by Meredig et al. [1].

    Notes
    -----
    This class is meant to be used in conjunction with
    ElementFractionAttributeGenerator and ValenceShellAttributeGenerator.

    To match the attributes from the Meredig et al. [1] paper, use all three
    attribute generators.

    References
    ----------
    .. [1] B. Meredig et al., "Combinatorial screening for new materials in
    unconstrained composition space with machine learning," Physical Review
    B, vol. 89, no. 9, Mar. 2014.

    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert feature headers here.
        feat_headers.append("mean_AtomicWeight")
        feat_headers.append("mean_Column")
        feat_headers.append("mean_Row")
        feat_headers.append("maxdiff_AtomicNumber")
        feat_headers.append("mean_AtomicNumber")
        feat_headers.append("maxdiff_CovalentRadius")
        feat_headers.append("mean_CovalentRadius")
        feat_headers.append("maxdiff_Electronegativity")
        feat_headers.append("mean_Electronegativity")
        feat_headers.append("mean_NsValence")
        feat_headers.append("mean_NpValence")
        feat_headers.append("mean_NdValence")
        feat_headers.append("mean_NfValence")

        # Load all property tables.
        mass = LookUpData.load_property("AtomicWeight")
        column = LookUpData.load_property("Column")
        row = LookUpData.load_property("Row")
        number = LookUpData.load_property("Number")
        radius = LookUpData.load_property("CovalentRadius")
        en = LookUpData.load_property("Electronegativity")
        s = LookUpData.load_property("NsValence")
        p = LookUpData.load_property("NpValence")
        d = LookUpData.load_property("NdValence")
        f = LookUpData.load_property("NfValence")

        for entry in entries:
            tmp_list = []
            elem_fractions = entry.get_element_fractions()
            elem_ids = entry.get_element_ids()
            tmp_mass = []
            tmp_column = []
            tmp_row = []
            tmp_number = []
            tmp_radius = []
            tmp_en = []
            tmp_s = []
            tmp_p = []
            tmp_d = []
            tmp_f = []
            for elem_id in elem_ids:
                tmp_mass.append(mass[elem_id])
                tmp_column.append(column[elem_id])
                tmp_row.append(row[elem_id])
                tmp_number.append(number[elem_id])
                tmp_radius.append(radius[elem_id])
                tmp_en.append(en[elem_id])
                tmp_s.append(s[elem_id])
                tmp_p.append(p[elem_id])
                tmp_d.append(d[elem_id])
                tmp_f.append(f[elem_id])

            tmp_list.append(np.average(tmp_mass, weights=elem_fractions))
            tmp_list.append(np.average(tmp_column, weights=elem_fractions))
            tmp_list.append(np.average(tmp_row, weights=elem_fractions))
            tmp_list.append(max(tmp_number) - min(tmp_number))
            tmp_list.append(np.average(tmp_number, weights=elem_fractions))
            tmp_list.append(max(tmp_radius) - min(tmp_radius))
            tmp_list.append(np.average(tmp_radius, weights=elem_fractions))
            tmp_list.append(max(tmp_en) - min(tmp_en))
            tmp_list.append(np.average(tmp_en, weights=elem_fractions))
            tmp_list.append(np.average(tmp_s, weights=elem_fractions))
            tmp_list.append(np.average(tmp_p, weights=elem_fractions))
            tmp_list.append(np.average(tmp_d, weights=elem_fractions))
            tmp_list.append(np.average(tmp_f, weights=elem_fractions))

            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import numpy as np
from mpi4py import MPI
import warnings
import multiprocessing
import nn_psgd
from ..utils import chunk

def train(X,Y,nneurons,input_act_funcs,validation_size=0.2,learn_rate=0.001,rms_decay=0.9,n_epochs=10000,
    batch_size=256,n_hist=20,n_check=50,threshold=0.1, print_level=1):
    """
    Main distributed memory function
    
    Parameters
    ----------
    All available parameters for nn_psgd - n_cores
    The number of cores will be directly passed to the mpirun command
            
    Returns
    -------
    trained_network: a list of dicts with trained weights and the activation functions from
    each node

    """

    # MPI
    comm=MPI.COMM_WORLD
    rank=comm.rank
    size=comm.size
    cpu_count = multiprocessing.cpu_count()
    cpu_count = comm.gather(cpu_count,root=0)
    if rank == 0:
        N = len(X)
        n_cores = sum(cpu_count)
        chunk_list= list( chunk(range(N),n_cores) )
        indices =[]
        for i,c in enumerate(cpu_count):
            indices = []
            for j in range(c):
                indices+=chunk_list.pop()
            if i!=0:
                comm.send(X[indices],dest=i, tag = 7)
                comm.send(Y[indices],dest=i, tag = 77)
            else:
                Xnew = X[indices]
                Ynew = Y[indices]
        X = Xnew
        Y = Ynew
    else:
        X = comm.recv(source=0, tag = 7)
        Y = comm.recv(source=0, tag = 77)

    trained_network =  nn_psgd.train(X,Y,nneurons=nneurons,
    input_act_funcs=input_act_funcs,learn_rate=learn_rate,rms_decay=rms_decay,
    n_epochs=n_epochs,batch_size=batch_size,n_cores=multiprocessing.cpu_count(),n_hist=n_hist,
    n_check=n_check,threshold=threshold, print_level=print_level)
    
    trained_network = comm.gather(trained_network,root=0)
    if rank==0:
        return trained_network

def output(X,nnets):
    """(nn_dsgd_output)
    User accessible output for neural network given trained weights.
    
    Parameters
    ----------
        X: array
            Input features
        
        nnets: list of dict
            A list of neural networks from each cluster. keys required weights and
             activation functions
    Returns
    -------
    predicted values in array type
    """
    #MPI
    comm = MPI.COMM_WORLD
    rank = comm.rank
    size = comm.size

    if rank == 0:
        results = []
        for nn in nnets:
            results+= [nn_psgd._output(X,nn['weights'],nn_psgd.act_funcs_from_string(nn['act_funcs'],len(nn['weights'])-1))]
        return results


import types
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData

class ValenceShellAttributeGenerator:
    """Class that generates attributes based on fraction of electrons in
    valence shell of constituent elements.

    Creates 4 features: [Composition-weighted mean # of electrons in the {s,p,
    d,f} shells]/[Mean # of Valence Electrons]

    Originally presented by: Meredig et al. [1].

    References
    ----------
    .. [1] B. Meredig et al., "Combinatorial screening for new materials in
    unconstrained composition space with machine learning," Physical Review
    B, vol. 89, no. 9, Mar. 2014.
    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        shell = ['s','p','d','f']
        n_valence = np.zeros((len(shell), len(LookUpData.element_names)))

        # Read in the data from tables and insert feature headers here.
        for i in range(len(shell)):
            s = shell[i]
            feat_headers.append("frac_"+s+"Valence")
            n_valence[i] = LookUpData.load_property("N"+s+"Valence")

        for entry in entries:
            elems = entry.get_element_ids()
            fracs = entry.get_element_fractions()
            sum_e = 0.0
            total_e = []
            for i in range(len(shell)):
                tmp_valence = []
                for elem_id in elems:
                    tmp_valence.append(n_valence[i][elem_id])

                # Fraction weighted average # of electrons in this shell.
                x = np.average(tmp_valence, weights=fracs)
                sum_e += x
                total_e.append(x)

            for i in range(len(total_e)):
                total_e[i] /= sum_e

            feat_values.append(total_e)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


"""
The cheml.chem module includes (please click on links adjacent to function names for more information):
    - CoulombMatrix: :func:`~cheml.chem.CoulombMatrix`
    - BagofBonds: :func:`~cheml.chem.BagofBonds`
    - RDKitFingerprint: :func:`~cheml.chem.RDKitFingerprint`
    - Dragon: :func:`~cheml.chem.Dragon`
    - DistanceMatrix: :func:`~cheml.chem.DistanceMatrix`
"""


from .CoulMat import CoulombMatrix
from .CoulMat import BagofBonds
from .RDKFP import RDKitFingerprint
from .Dragon import Dragon
from .DistMat import DistanceMatrix


__all__ = [
    'CoulombMatrix',
    'BagofBonds',
    'RDKitFingerprint',
    'Dragon',
    'DistanceMatrix',
]




import pandas as pd
import numpy as np
import types
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry

class StructuralHeterogeneityAttributeGenerator:
    """Class to compute attributes based on heterogeneity in structure.

    Measures variance in bond lengths (both for a single atom and between
    different atoms) and atomic volumes. Also considers the number of unique
    coordination polyhedron shapes.

    Bond lengths, atomic volumes, and coordination polyhedra are based on the
    Voronoi tessellation of the structure.

    Notes
    -----
    Current attributes:
    1. Mean absolute deviation in average bond length for each atom, normalized
    by mean for all atoms.
    2. Minimum in average bond length, normalized by mean for all atoms.
    3. Maximum in average bond length, normalized by mean for all atoms.
    4. Mean bond length variance between bonds across all atom.
    5. Mean absolute deviation in bond length variance.
    6. Minimum bond length variance.
    7. Maximum bond length variance.
    8. Mean absolute deviation in atomic volume, normalized by mean atomic
    volume.

    Here, bond length variation for a single atom is defined as:
    .. math:: \hat{l} = <l_i - \bar{l}>
    where :math: `l_i` is the distance between an atom and one of its
    neighbors.

    """

    def generate_features(self, entries):
        """Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Crystal structures for which features are to be generated. A list
            of CrystalStructureEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CrystalStructureEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")
        elif (entries and not isinstance(entries[0], CrystalStructureEntry)):
            raise ValueError("Argument should be of type list of "
                             "CrystalStructureEntry's")

        # Insert header names here.
        feat_headers.append("var_MeanBondLength")
        feat_headers.append("min_MeanBondLength")
        feat_headers.append("max_MeanBondLength")
        feat_headers.append("mean_BondLengthVariation")
        feat_headers.append("var_BondLengthVariation")
        feat_headers.append("min_BondLengthVariation")
        feat_headers.append("max_BondLengthVariation")
        feat_headers.append("var_CellVolume")

        l_fh = len(feat_headers)
        # Generate features for each entry.
        for entry in entries:
            tmp_list = []
            # Get the Voronoi tessellation.
            try:
                voro = entry.compute_voronoi_tessellation()
            except Exception:
                tmp_list = [np.nan] * l_fh # If tessellation fails.
                feat_values.append(tmp_list)
                continue

            # Bond length features.
            # Variation between cells.
            mean_bond_lengths = voro.mean_bond_lengths()
            mean_bond_lengths /= np.mean(mean_bond_lengths) # Normalize bond
            # lengths.
            m = np.mean(mean_bond_lengths)
            tmp_list.append(np.mean([abs(x - m) for x in mean_bond_lengths]))
            tmp_list.append(np.min(mean_bond_lengths))
            tmp_list.append(np.max(mean_bond_lengths))

            # Variation within a single cell.
            mean_bond_lengths = voro.mean_bond_lengths() # Recompute bond
            # lengths.
            bond_length_variation = voro.bond_length_variance(
                mean_bond_lengths)
            # Normalize bond length variation by mean bond length of each cell.
            bond_length_variation /= mean_bond_lengths
            m = np.mean(bond_length_variation)
            tmp_list.append(m)
            tmp_list.append(np.mean([abs(x - m) for x in
                                     bond_length_variation]))
            tmp_list.append(np.min(bond_length_variation))
            tmp_list.append(np.max(bond_length_variation))

            # Cell volume / shape features.
            tmp_list.append(voro.volume_variance() * entry.get_structure(
            ).n_atoms() / entry.get_structure().volume())

            feat_values.append(tmp_list)

        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import unittest
import numpy as np
import numpy.testing as np_tst
from ....vassal.util.VectorCombinationComputer import VectorCombinationComputer
import math

class testVectorCombinationComputer(unittest.TestCase):
    def setUp(self):
        self.input = list(np.identity(3))

    def tearDown(self):
        self.input = None

    def test_combination(self):
        computer = VectorCombinationComputer(self.input, 1)

        v = np.array([1, 0, 0])
        np_tst.assert_array_almost_equal(v, computer.compute_vector(v))
        v = np.array([0, 1, 0])
        np_tst.assert_array_almost_equal(v, computer.compute_vector(v))
        v = np.array([0, 0, 1])
        np_tst.assert_array_almost_equal(v, computer.compute_vector(v))
        v = np.array([-2, 5, 3])
        np_tst.assert_array_almost_equal(v, computer.compute_vector(v))

    def test_collection(self):
        computer = VectorCombinationComputer(self.input, 1)
        self.assertEquals(7, len(computer.get_vectors()))

        computer = VectorCombinationComputer(self.input, math.sqrt(2) * 1.001)
        self.assertEquals(1 + 6 + 12, len(computer.get_vectors()))

        computer = VectorCombinationComputer(self.input, math.sqrt(2) *
                                             1.001, include_zero=False)
        self.assertEquals(6 + 12, len(computer.get_vectors()))

        computer = VectorCombinationComputer(self.input, 8)
        self.assertEquals(len(computer.get_vectors()),
                          len(computer.get_vectors()))

    def test_collection_oblique(self):
        self.input[1][0] = 8

        computer = VectorCombinationComputer(self.input, 1)
        self.assertEquals(7, len(computer.get_vectors()))

        computer = VectorCombinationComputer(self.input, math.sqrt(2) * 1.001)
        self.assertEquals(1 + 6 + 12, len(computer.get_vectors()))


import copy
import warnings

import numpy as np
import pandas as pd

from .database import sklearn_db
from .database import cheml_db
from .database import pandas_db


# todo: decorate some of the steps in the wrapeprs. e.g. sending out ouputs by finding all the connected edges in the graph
# todo: use Input and Output classes to handle inputs and outputs

class BASE(object):
    """
    Do not instantiate this class
    """
    def __init__(self, Base, parameters, iblock, Task, Function, Host):
        self.Base = Base
        self.parameters = parameters
        self.iblock = iblock
        self.Task = Task
        self.Function = Function
        self.Host = Host

    def run(self):
        self.IO()
        self.Receive()
        self.fit()

    def IO(self):
        if self.Host == 'sklearn':
            self.metadata = getattr(sklearn_db, self.Function)()
        elif self.Host == 'cheml':
            self.metadata = getattr(cheml_db, self.Function)()
        elif self.Host == 'pandas':
            self.metadata = getattr(pandas_db, self.Function)()
        self.inputs = {i:copy.deepcopy(vars(self.metadata.Inputs)[i]) for i in vars(self.metadata.Inputs).keys() if
                       i not in ('__module__','__doc__')}
        self.outputs = {i:copy.deepcopy(vars(self.metadata.Outputs)[i]) for i in vars(self.metadata.Outputs).keys() if
                   i not in ('__module__', '__doc__')}
        # self.wparams = {i:copy.deepcopy(vars(self.metadata.WParameters)[i]) for i in vars(self.metadata.WParameters).keys() if
        #             i not in ('__module__', '__doc__')}

    def Receive(self):
        recv = [edge for edge in self.Base.graph if edge[2] == self.iblock]
        self.Base.graph = tuple([edge for edge in self.Base.graph if edge[2] != self.iblock])
        # check received tokens to: (1) be a legal input, and (2) be unique.
        count = {token: 0 for token in self.inputs}
        for edge in recv:
            if edge[3] in self.inputs:
                count[edge[3]] += 1
                if count[edge[3]] > 1:
                    msg = '@Task #%i(%s): only one input per each available input token can be received.' % (
                        self.iblock + 1, self.Task)
                    raise IOError(msg)
            else:
                msg = "@Task #%i(%s): received a non valid input token '%s', sent by block #%i" % (
                    self.iblock + 1, self.Task, edge[3], edge[0] + 1)
                raise IOError(msg)
        for edge in recv:
            key = edge[0:2]
            if key in self.Base.send:
                if self.Base.send[key].count > 0:
                    value = self.Base.send[key].value
                    # Todo: add an option to deepcopy(value)
                    if str(type(value)) in self.inputs[edge[3]].types or \
                               len(self.inputs[edge[3]].types)==0:
                        self.inputs[edge[3]].value = value
                        self.inputs[edge[3]].fro = self.Base.send[key].fro
                        self.Base.send[key].count -= 1
                    else:
                        msg = "@Task #%i(%s): The input token '%s' doesn't support the received format"  % (
                            self.iblock + 1, self.Task, edge[3])
                        raise IOError(msg)
                if self.Base.send[key].count == 0:
                    del self.Base.send[key]
            else:
                msg = "@Task #%i(%s): no output has been sent to the input token '%s'" % (
                    self.iblock + 1, self.Task, edge[3])
                raise IOError(msg)

    def Send(self):
        order = [edge[1] for edge in self.Base.graph if edge[0] == self.iblock]
        for token in set(order):
            if token in self.outputs:
                if self.outputs[token].value is not None:
                    self.outputs[token].count = order.count(token)
                    self.Base.send[(self.iblock, token)] = self.outputs[token]
                else:
                    msg = "@Task #%i(%s): not allowed to send out empty objects '%s'" % (
                        self.iblock + 1, self.Task, token)
                    warnings.warn(msg)
            else:
                msg = "@Task #%i(%s): not a valid output token '%s'" % (self.iblock + 1, self.Task, token)
                raise NameError(msg)

    def required(self, token, req=False):
        """
        Tasks:
            - check if input token is required

        :param token: string, name of the input
        :param req: Boolean, optional (default = False)

        """
        if self.inputs[token].value is None:
            if req:
                msg = "@Task #%i(%s): The input '%s' is required." \
                      % (self.iblock + 1, self.Task, token)
                raise IOError(msg)

    def paramFROMinput(self):
        for param in self.parameters:
            if isinstance(self.parameters[param], str):
                if self.parameters[param][0]=='@':
                    token = self.parameters[param][1:].strip()
                    if token in self.inputs:
                        self.parameters[param] = self.inputs[token].value
                    else:
                        msg = "@Task #%i(%s): assigned an unknown token name - %s - to the parameter - %s - " \
                              % (self.iblock + 1, self.Task, token, param)
                        raise IOError(msg)

    def set_value(self,token,value):
        self.outputs[token].fro = (self.iblock,self.Host,self.Function)
        if str(type(value)) in self.outputs[token].types or \
                    len(self.outputs[token].types)==0:
            self.outputs[token].value = value
        else:
            msg = "@Task #%i(%s): The output token '%s' doesn't support the type"  % (
                self.iblock, self.Host, token)
            raise IOError(msg)

    def import_sklearn(self):
        try:
            exec ("from %s.%s import %s" % (self.metadata.modules[0], self.metadata.modules[1], self.Function))
            submodule = getattr(__import__(self.metadata.modules[0]), self.metadata.modules[1])
            F = getattr(submodule, self.Function)
            api = F(**self.parameters)
        except Exception as err:
            msg = '@Task #%i(%s): ' % (self.iblock + 1, self.Task) + type(err).__name__ + ': ' + err.message
            raise TypeError(msg)
        return api

    def Fit_sklearn(self):
        self.paramFROMinput()
        if 'track_header' in self.parameters:
            self.header = self.parameters.pop('track_header')
        else:
            self.header = True
        if 'func_method' in self.parameters:
            self.method = self.parameters.pop('func_method')
        else:
            self.method = None

        available_methods = self.metadata.WParameters.func_method.options
        if self.method not in available_methods:
            msg = "@Task #%i(%s): The method '%s' is not available for the function '%s'." % (
                self.iblock, self.Task,self.method,self.Function)
            raise NameError(msg)

        # methods: fit, predict, None for regression
        # methods: fit_transform, transform, inverse_transform, None for transformers
        if self.method == None:
            api = self.import_sklearn()
            self.set_value('api', api)
        elif self.method == 'fit_transform':
            api = self.import_sklearn()
            self.required('df', req=True)
            df = self.inputs['df'].value
            df = api.fit_transform(df)
            self.set_value('api', api)
            self.set_value('df', pd.DataFrame(df))
        elif self.method == 'transform':
            self.required('df', req=True)
            self.required('api', req=True)
            df = self.inputs['df'].value
            api = self.inputs['api'].value
            df = api.transform(df)
            self.set_value('api', api)
            self.set_value('df', pd.DataFrame(df))
        elif self.method == 'inverse_transform':
            self.required('df', req=True)
            self.required('api', req=True)
            df = self.inputs['df'].value
            api = self.inputs['api'].value
            df = api.inverse_transform(df)
            self.set_value('api', api)
            self.set_value('df', pd.DataFrame(df))
        elif self.method == 'fit':
            api = self.import_sklearn()
            self.required('dfx', req=True)
            self.required('dfy', req=True)
            dfx = self.inputs['dfx'].value
            dfy = self.inputs['dfy'].value
            api.fit(dfx,dfy)
            dfy_predict = api.predict(dfx)
            self.set_value('api', api)
            self.set_value('dfy_predict', pd.DataFrame(dfy_predict))
        elif self.method == 'predict':
            self.required('dfx', req=True)
            self.required('api', req=True)
            dfx = self.inputs['dfx'].value
            api = self.inputs['api'].value
            dfy_predict = api.predict(dfx)
            self.set_value('api', api)
            self.set_value('dfy_predict', pd.DataFrame(dfy_predict))

    def _dim_check(self, token, X, ndim):
        if (X.ndim == ndim < 3):
            pass
        elif (X.ndim == 1) and (ndim == 2):
            X = X.reshape(-1,1)
        elif (X.ndim == 2) and (X.shape[1] == 1) and (ndim == 1):
            X = X.ravel()
        else:
            msg = "@Task #%i(%s): the %s is not or can not be converted to %i dimensional " \
                  % (self.iblock + 1, self.Task, token, ndim)
            raise IOError(msg)
        return X

    def data_check(self, token, X, ndim=2, n0=None, n1=None, format_out='df'):
        """
        Tasks:
            - check the dimension and size of input
            - change the format from numpy array to pandas data frame or vice versa

        :param X: numpy.ndarray or pandas.DataFrame
            input data
        :param token: string
            name of input (e.g. training input)
        :param ndim: integer, optional (default=2)
            X.ndim; valid digits are 1 and 2
        :param n0: int
            number of data entries
        :param n1: int
            number of features
        :param format_out: string ('df' or 'ar'), optional (default = 'df')

        :return input data converted to array or dataframe
        :return the header of dataframe
            if input data is not a dataframe return None
        """
        if isinstance(X, pd.DataFrame):
            if format_out == 'ar':
                # print '%s.ndim:'%token, X.values.ndim, "; changing to %i-dimension ..." %ndim
                header = X.columns
                X = self._dim_check(token, X.values, ndim)
            else:
                header = X.columns
            # if not np.can_cast(X.dtypes, np.float, casting='same_kind'):
            #     msg = "@Task #%i(%s): %s cannot be cast to floats" \
            #           % (self.iblock + 1, self.Task, token)
            #     raise Exception(msg)
        elif isinstance(X, np.ndarray):
            if format_out == 'df':
                X = pd.DataFrame(X)
                header = None
            else:
                header = None
                X = self._dim_check(token, X, ndim)
        else:
            msg = "@Task #%i(%s): %s needs to be either pandas dataframe or numpy array" \
                  % (self.iblock + 1, self.Task, token)
            raise Exception(msg)

        if n0 and X.shape[0] != n0:
            msg = "@Task #%i(%s): %s has an inconsistant number of data entries" \
                  % (self.iblock + 1, self.Task, token)
            raise Exception(msg)
        if n1 and X.shape[1] != n1:
            msg = "@Task #%i(%s): %s has an inconsistant number of feature entries" \
                  % (self.iblock + 1, self.Task, token)
            raise Exception(msg)
        return X, header #X.astype(float), header

class LIBRARY(object):
    """
    Do not instantiate this class
    """
    def references(self,host,function):
        from ..chem.magpie_python import __all__ as magpie_all
        from ..visualization import __all__ as matplotlib_all

        # numpy
        ref_g = "https://github.com/numpy/numpy"
        ref_p = "Travis E, Oliphant. A guide to NumPy, USA: Trelgol Publishing, (2006)."
        self.refs['NumPy'] = {'github': ref_g, 'paper': ref_p}
        # pandas
        ref_g = "https://github.com/pandas-dev/pandas"
        ref_p = "Wes McKinney. Data Structures for Statistical Computing in Python, Proceedings of the 9th Python in Science Conference, 51-56 (2010)"
        self.refs['Pandas'] = {'github': ref_g, 'paper': ref_p}

        if host == 'sklearn':
            ref_g = "https://github.com/scikit-learn/scikit-learn"
            ref_p = "Scikit-learn: Machine Learning in Python, Pedregosa et al., JMLR 12, pp. 2825-2830, 2011."
            self.refs['scikit-learn'] = {'github':ref_g, 'paper':ref_p}
        elif host == 'cheml':
            ref_g = "https://github.com/hachmannlab/ChemML"
            ref_p = "Haghighatlari M, Subramanian R, Urala B, Vishwakarma G, Sonpal A, Chen P, Setlur S, Hachmann J (2017), ChemML: A machine learning and informatics program suite for the chemical and materials sciences, https://github.com/hachmannlab/ChemML"
            self.refs['ChemML'] =  {'github': ref_g, 'paper': ref_p}
            if function == 'RDKitFingerprint':
                ref_g = "https://github.com/rdkit"
                ref_p = "no publication"
                self.refs['rdkit'] = {'github': ref_g, 'paper': ref_p}
            elif function == 'Dragon':
                ref_g = "http://www.talete.mi.it/products/dragon_description.htm"
                ref_p = "R. Todeschini,  V. Consonni,  R. Mannhold,H.  Kubinyi,  H.  Timmerman,  Handbook  ofMolecular Descriptors, Wiley-VCH, 2000."
                self.refs['Dragon'] = {'url': ref_g, 'paper': ref_p}
            elif function == 'CoulombMatrix':
                ref_g = "no software package"
                ref_p = "M.  Rupp,  A.  Tkatchenko,  K.-R.  Muller,O.  A.  von  Lilienfeld,   Fast  and  Accu-rate  Modeling  of  Molecular  AtomizationEnergies  with  Machine  Learning,   Physi-cal  Review  Letters  108  (5)  (2012)  058301.doi:10.1103/PhysRevLett.108.058301"
                self.refs['CoulombMatrix'] = {'url': ref_g, 'paper': ref_p}
            elif function == 'BagofBonds':
                ref_g = "no software package"
                ref_p = "Hansen, K.; Biegler, F.; Ramakrishnan, R.; Pronobis, W.; von Lilienfeld, O. A.; Muller, K.-R.; Tkatchenko, A. Machine Learning Predictions of Molecular Properties: Accurate Many-Body Potentials and Nonlocality in Chemical Space J. Phys. Chem. Lett. 2015, 6, 2326 2331, DOI: 10.1021/acs.jpclett.5b00831"
                self.refs['BagofBonds'] = {'url': ref_g, 'paper': ref_p}
            elif function in ['MLP','MLP_sklearn']:
                ref_g = "https://github.com/fchollet/keras"
                ref_p = "@misc{chollet2015keras,title={Keras},author={Chollet, Fran\c{c}ois and others},year={2015},publisher={GitHub},howpublished={\url{https://github.com/keras-team/keras}},}"
                self.refs['keras'] = {'url': ref_g, 'paper': ref_p}
            elif function in ['GA_DEAP']:
                ref_g = "https://github.com/deap/deap"
                ref_p = """@article{DEAP_JMLR2012,
                                author    = " F\'elix-Antoine Fortin and Fran\c{c}ois-Michel {De Rainville} and Marc-Andr\'e Gardner and Marc Parizeau and Christian Gagn\'e ",
                                title     = { {DEAP}: Evolutionary Algorithms Made Easy },
                                pages    = { 2171--2175 },
                                volume    = { 13 },
                                month     = { jul },
                                year      = { 2012 },
                                journal   = { Journal of Machine Learning Research }
                            }"""
                self.refs['deap'] = {'url': ref_g, 'paper': ref_p}
            elif function in magpie_all:
                ref_g = "https://bitbucket.org/wolverton/magpie"
                ref_p = "L. Ward, A. Agrawal, A. Choudhary, and C. Wolverton, A general-purpose machine learning framework for predicting properties of inorganic materials, npj Computational Materials, vol. 2, no. 1, Aug. 2016."
                self.refs['magpie'] = {'url': ref_g, 'paper': ref_p}
            elif function in matplotlib_all:
                ref_g = "https://bitbucket.org/wolverton/magpie"
                ref_p = """@Article{Hunter:2007,
                              Author    = {Hunter, J. D.},
                              Title     = {Matplotlib: A 2D graphics environment},
                              Journal   = {Computing In Science \& Engineering},
                              Volume    = {9},
                              Number    = {3},
                              Pages     = {90--95},
                              abstract  = {Matplotlib is a 2D graphics package used for Python
                              for application development, interactive scripting, and
                              publication-quality image generation across user
                              interfaces and operating systems.},
                              publisher = {IEEE COMPUTER SOC},
                              doi       = {10.1109/MCSE.2007.55},
                              year      = 2007
                        }"""
                self.refs['matplotlib'] = {'url': ref_g, 'paper': ref_p}
        elif host == 'tf':
            ref_g = "https://github.com/tensorflow/tensorflow"
            ref_p = "M. Abadi,  P. Barham,  J. Chen,  Z. Chen,A. Davis, J. Dean, M. Devin, S. Ghemawat,G. Irving, M. Isard, M. Kudlur, J. Levenberg,R. Monga, S. Moore, D. G. Murray, B. Steiner,P. Tucker, V. Vasudevan, P. Warden, M. Wicke,Y. Yu, X. Zheng, Tensorflow:  A system forlarge-scale machine learning, in: 12th USENIXSymposium on Operating Systems Design andImplementation (OSDI 16), USENIX Associa-tion, GA, 2016, pp. 265-283"
            self.refs['tensorflow'] = {'github': ref_g, 'paper': ref_p}

    def _save_references(self):
        with open(self.Base.output_directory+'/citation.txt','w') as file:
            for module in self.refs:
                file.write(module+':\n')
                for source in self.refs[module]:
                    file.write('    '+source+': '+self.refs[module][source]+'\n')
                file.write('\n')


import pandas as pd
import numpy as np


class CoulombMatrix(object):
    """ (CoulombMatrix)
    The implementation of coulomb matrix by Matthias Rupp et al 2012, PRL (3 different representations).

    Parameters
    ----------
    CMtype: string, optional (default='SC')
        The coulomb matrix type, one of the following types:
            * 'Unsorted_Matrix' or 'UM'
            * 'Unsorted_Triangular' or 'UT'
            * 'Eigenspectrum' or 'E' 
            * 'Sorted_Coulomb' or 'SC'
            * 'Random_Coulomb' or 'RC'
    max_n_atoms: integer or 'auto', optional (default = 'auto')
        maximum number of atoms in a molecule. if 'auto' we find it based on the input molecules.
    nPerm: integer, optional (default = 3)
        Number of permutation of coulomb matrix per molecule for Random_Coulomb (RC) 
        type of representation.
    const: float, optional (default = 1)
            The constant value for coordinates unit conversion to atomic unit
            example: atomic unit -> const=1, Angstrom -> const=0.529
            const/|Ri-Rj|, which denominator is the euclidean distance
            between atoms i and j
    """
    def __init__(self, CMtype='SC', max_n_atoms = 'auto', nPerm=3, const=1):
        self.CMtype = CMtype
        self.max_n_atoms = max_n_atoms
        self.nPerm = nPerm
        self.const = const

    def __cal_coul_mat(self, mol):
        cm = []
        for i in xrange(len(mol)):
            vect = []
            for k in range(0,i):
                vect.append(cm[k][i])
            for j in range(i,len(mol)):
                if i==j:
                    vect.append(0.5*mol[i,0]**2.4)
                else:
                    vect.append((mol[i,0]*mol[j,0]*self.const)/np.linalg.norm(mol[i,1:]-mol[j,1:]))
            for m in range(len(mol),self.max_n_atoms):
                vect.append(0.0)
            cm.append(vect)
        for m in range(len(mol),self.max_n_atoms):
            cm.append([0]*self.max_n_atoms)
        return np.array(cm) #shape nAtoms*nAtoms

    def represent(self, molecules):
        """
        provides coulomb matrix representation for input molecules.

        Parameters
        ----------
        molecules: dictionary, numpy array or pandas data frame
            if dictionary: output of cheml.initialization.XYZreader
            if numpy array: an array or list of molecules, which each molecule has a shape of (number of atoms, 4)
            if pandas dataframe : only one molecule with shape (number of atoms, 4).
            The four columns are nuclear charge, x-corrdinate, y-coordinate, z-coordinate for each atom, respectively.

        Returns
        -------
        pandas DataFrame
            shape of Unsorted_Matrix (UM): (n_molecules, max_n_atoms**2)
            shape of Unsorted_Triangular (UT): (n_molecules, max_n_atoms*(max_n_atoms+1)/2)
            shape of eigenspectrums (E): (n_molecules, max_n_atoms)
            shape of Sorted_Coulomb (SC): (n_molecules, max_n_atoms*(max_n_atoms+1)/2)
            shape of Random_Coulomb (RC): (n_molecules, nPerm * max_n_atoms * (max_n_atoms+1)/2)

        """
        if isinstance(molecules, dict):
            molecules = np.array([molecules[i]['mol'] for i in xrange(1, len(molecules)+1)])
        elif isinstance(molecules, pd.DataFrame):
            # only one molecule
            molecules = np.array([molecules.values])

        self.n_molecules = len(molecules)
        if self.max_n_atoms == 'auto':
            self.max_n_atoms = max([len(m) for m in molecules])
        if self.CMtype == "Unsorted_Matrix" or self.CMtype == 'UM':
            cms = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol)
                cms = np.append(cms, cm.ravel())
            cms = cms.reshape(self.n_molecules, self.max_n_atoms**2)
            cms = pd.DataFrame(cms)
            return cms

        elif self.CMtype == "Unsorted_Triangular" or self.CMtype == 'UT':
            cms = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol)
                cms = np.append(cms, cm[np.tril_indices(self.max_n_atoms)])
            cms = cms.reshape(self.n_molecules, self.max_n_atoms*(self.max_n_atoms+1)/2)
            cms = pd.DataFrame(cms)
            return cms

        elif self.CMtype == 'Eigenspectrum' or self.CMtype == 'E':
            eigenspectrums = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol) # Check the constant value for unit conversion; atomic unit -> 1 , Angstrom -> 0.529
                eig = np.linalg.eigvals(cm)
                eig[::-1].sort()
                eigenspectrums = np.append(eigenspectrums,eig)
            eigenspectrums = eigenspectrums.reshape(self.n_molecules, self.max_n_atoms)
            eigenspectrums = pd.DataFrame(eigenspectrums)
            return eigenspectrums
            
        elif self.CMtype == 'Sorted_Coulomb' or self.CMtype == 'SC':
            sorted_cm = np.array([])
            for mol in molecules:
                cm = self.__cal_coul_mat(mol)
                lambdas = np.linalg.norm(cm,2,1)
                sort_indices = np.argsort(lambdas)[::-1]
                cm = cm[:,sort_indices][sort_indices,:]
                # sorted_cm.append(cm)
                sorted_cm = np.append(sorted_cm, cm[np.tril_indices(self.max_n_atoms)]) # lower-triangular
            sorted_cm = sorted_cm.reshape(self.n_molecules, self.max_n_atoms*(self.max_n_atoms+1)/2)
            sorted_cm = pd.DataFrame(sorted_cm)
            return sorted_cm
                    
        elif self.CMtype == 'Random_Coulomb' or self.CMtype == 'RC':
            random_cm = np.array([])
            for nmol,mol in enumerate(molecules):
                cm = self.__cal_coul_mat(mol)
                lambdas = np.linalg.norm(cm,2,1)
                sort_indices = np.argsort(lambdas)[::-1]
                cm = cm[:,sort_indices][sort_indices,:]
                cm_perms = []
                for l in range(self.nPerm):
                    mask = np.random.permutation(self.max_n_atoms)
                    cm_mask = cm[:,mask][mask,:]
                    # cm_perms.append(cm_mask)
                    cm_perms += list(cm_mask[np.tril_indices(self.max_n_atoms)]) # lower-triangular
                random_cm = np.append(random_cm, cm_perms)
            random_cm = random_cm.reshape(self.n_molecules, self.nPerm*self.max_n_atoms*(self.max_n_atoms+1)/2)
            return pd.DataFrame(random_cm)

class BagofBonds(object):
    """ (BagofBonds)
    The implementation of bag of bonds version of coulomb matrix by katja Hansen et al 2015, JPCL.

    Parameters
    ----------
    const: float, optional (default = 1.0)
            The constant value for coordinates unit conversion to atomic unit
            if const=1.0, returns atomic unit
            if const=0.529, returns Angstrom

            const/|Ri-Rj|, which denominator is the euclidean distance
            between two atoms

    Attributes
    ----------
    header: list of header for the bag of bonds data frame
        contains one nuclear charge (represents single atom) or a tuple of two nuclear charges (represents a bond)

    Examples
    --------
    >>> from cheml.datasets import load_xyz_polarizability
    >>> from cheml.chem import BagofBonds

    >>> coordinates, y = load_xyz_polarizability()
    >>> bob = BagofBonds(const= 1.0)
    >>> X = bob.represent(coordinates)
    """
    def __init__(self, const=1.0):
        self.const = const

    def represent(self, molecules):
        """
        provides bag of bonds representation for input molecules.

        Parameters
        ----------
        molecules: dictionary,numpy array  or pandas data frame
            if dictionary: output of cheml.initialization.XYZreader
            if numpy array: array or list of molecules, which each molecule is an array with shape (number of atoms, 4).
            if pandas dataframe : only one molecule with shape (number of atoms, 4).
            The four columns are nuclear charge, x-corrdinate, y-coordinate, z-coordinate for each atom, respectively.

        Returns
        -------
        pandas data frame, shape: (n_molecules, max_length_of_combinations)

        """
        if isinstance(molecules, dict):
            molecules = np.array([molecules[i]['mol'] for i in xrange(1,len(molecules)+1)])
        elif isinstance(molecules, pd.DataFrame):
            molecules = molecules.values

        BBs_matrix = [] # list of dictionaries for each molecule
        all_keys = {}   # dictionary of unique keys and their maximum length
        for nmol,mol in enumerate(molecules):
            bags = {}
            for i in xrange(len(mol)):
                for j in xrange(i,len(mol)):
                    if i==j:
                        key = (mol[i,0], mol[i,0])
                        Fc = 0.5 * mol[i, 0] ** 2.4
                        if key in bags:
                            bags[key].append(Fc)
                        else:
                            bags[key] = [Fc]
                    else:
                        key = (max(mol[i,0],mol[j,0]),min(mol[i,0],mol[j,0]))
                        Fc = (mol[i,0]*mol[j,0]*self.const) / np.linalg.norm(mol[i,1:]-mol[j,1:])
                        if key in bags:
                            bags[key].append(Fc)
                        else:
                            bags[key] = [Fc]
            for key in bags:
                if key in all_keys:
                    all_keys[key] = max(all_keys[key], len(bags[key]))
                else:
                    all_keys[key] = len(bags[key])
            BBs_matrix.append(bags)

        # extend bags with zero
        for i, bags in enumerate(BBs_matrix):
            for key in all_keys:
                if key in bags:
                    diff = all_keys[key] - len(bags[key])
                    BBs_matrix[i][key] = sorted(BBs_matrix[i][key], reverse=True) + diff*[0]
                else:
                    BBs_matrix[i][key] = all_keys[key]*[0]
        df = pd.DataFrame(BBs_matrix)
        del BBs_matrix
        order_headers = list(df.columns)
        output = pd.DataFrame(list(df.sum(1)))
        del df
        self.headers = []
        for key in order_headers:
            if key[0]==key[1]:
                k = key[0]
            else:
                k = key
            self.headers += all_keys[key] * [k]
        return output


import unittest
from itertools import permutations
import numpy.testing as np
import os
from ....data.materials.CompositionEntry import CompositionEntry

class testCompositionEntry(unittest.TestCase):
    def test_parsing(self):
        entry = CompositionEntry(composition="Fe")
        self.assertEquals(1, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0, entry.get_element_fraction(name="Fe"),
                                delta=1e-6)

        entry = CompositionEntry(composition="FeO0")
        self.assertEquals(1, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0, entry.get_element_fraction(name="Fe"),
                                delta=1e-6)

        entry = CompositionEntry(composition="FeCl3")
        self.assertEquals(2, len(entry.get_element_ids()))
        self.assertAlmostEquals(0.25, entry.get_element_fraction(name="Fe"),
                                delta=1e-6)
        self.assertAlmostEquals(0.75, entry.get_element_fraction(name="Cl"),
                                delta=1e-6)

        entry = CompositionEntry(composition="Fe1Cl_3")
        self.assertEquals(2, len(entry.get_element_ids()))
        self.assertAlmostEquals(0.25, entry.get_element_fraction(name="Fe"),
                                delta=1e-6)
        self.assertAlmostEquals(0.75, entry.get_element_fraction(name="Cl"),
                                delta=1e-6)

        entry = CompositionEntry(composition="FeCl_3")
        self.assertEquals(2, len(entry.get_element_ids()))
        self.assertAlmostEquals(0.25, entry.get_element_fraction(name="Fe"),
                                delta=1e-6)
        self.assertAlmostEquals(0.75, entry.get_element_fraction(name="Cl"),
                                delta=1e-6)

        entry = CompositionEntry(composition="FeClCl2")
        self.assertEquals(2, len(entry.get_element_ids()))
        self.assertAlmostEquals(0.25, entry.get_element_fraction(name="Fe"),
                                delta=1e-6)
        self.assertAlmostEquals(0.75, entry.get_element_fraction(name="Cl"),
                                delta=1e-6)

        entry = CompositionEntry(composition="Ca(NO3)2")
        self.assertEquals(3, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 9, entry.get_element_fraction(name="Ca"),
                                delta=1e-6)
        self.assertAlmostEquals(2.0 / 9, entry.get_element_fraction(name="N"),
                                delta=1e-6)
        self.assertAlmostEquals(6.0 / 9, entry.get_element_fraction(name="O"),
                                delta=1e-6)

        entry = CompositionEntry(composition="Ca(N[O]3)2")
        self.assertEquals(3, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 9, entry.get_element_fraction(name="Ca"),
                                delta=1e-6)
        self.assertAlmostEquals(2.0 / 9, entry.get_element_fraction(name="N"),
                                delta=1e-6)
        self.assertAlmostEquals(6.0 / 9, entry.get_element_fraction(name="O"),
                                delta=1e-6)

        entry = CompositionEntry(composition="Ca(N(O1.5)2)2")
        self.assertEquals(3, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 9, entry.get_element_fraction(name="Ca"),
                                delta=1e-6)
        self.assertAlmostEquals(2.0 / 9, entry.get_element_fraction(name="N"),
                                delta=1e-6)
        self.assertAlmostEquals(6.0 / 9, entry.get_element_fraction(name="O"),
                                delta=1e-6)

        entry = CompositionEntry(composition="Ca{N{O1.5}2}2")
        self.assertEquals(3, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 9, entry.get_element_fraction(name="Ca"),
                                delta=1e-6)
        self.assertAlmostEquals(2.0 / 9, entry.get_element_fraction(name="N"),
                                delta=1e-6)
        self.assertAlmostEquals(6.0 / 9, entry.get_element_fraction(name="O"),
                                delta=1e-6)

        entry = CompositionEntry(composition="CaO-0.01Ni")
        self.assertEquals(3, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 2.01, entry.get_element_fraction(
            name="Ca"), delta=1e-6)
        self.assertAlmostEquals(0.01 / 2.01, entry.get_element_fraction(
            name="Ni"), delta=1e-6)
        self.assertAlmostEquals(1.0 / 2.01, entry.get_element_fraction(
            name="O"), delta=1e-6)

        entry = CompositionEntry(composition="CaO"+str(chr(183))+"0.01Ni")
        self.assertEquals(3, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 2.01, entry.get_element_fraction(
            name="Ca"), delta=1e-6)
        self.assertAlmostEquals(0.01 / 2.01, entry.get_element_fraction(
            name="Ni"), delta=1e-6)
        self.assertAlmostEquals(1.0 / 2.01, entry.get_element_fraction(
            name="O"), delta=1e-6)

        entry = CompositionEntry(composition="Ca(N(O1.5)2)2-2H2O")
        self.assertEquals(4, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 15, entry.get_element_fraction(name="Ca"),
                                delta=1e-6)
        self.assertAlmostEquals(2.0 / 15, entry.get_element_fraction(name="N"),
                                delta=1e-6)
        self.assertAlmostEquals(8.0 / 15, entry.get_element_fraction(name="O"),
                                delta=1e-6)
        self.assertAlmostEquals(4.0 / 15, entry.get_element_fraction(name="H"),
                                delta=1e-6)

        entry = CompositionEntry(composition="Ca(N(O1.5)2)2-2.1(H)2O")
        self.assertEquals(4, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 15.3, entry.get_element_fraction(
            name="Ca"), delta=1e-6)
        self.assertAlmostEquals(2.0 / 15.3, entry.get_element_fraction(
            name="N"), delta=1e-6)
        self.assertAlmostEquals(8.1 / 15.3, entry.get_element_fraction(
            name="O"), delta=1e-6)
        self.assertAlmostEquals(4.2 / 15.3, entry.get_element_fraction(
            name="H"), delta=1e-6)

        entry = CompositionEntry(composition="{[("
                            "Fe0.6Co0.4)0.75B0.2Si0.05]0.96Nb0.04}96Cr4")
        self.assertEquals(6, len(entry.get_element_ids()))
        self.assertAlmostEquals(0.41472, entry.get_element_fraction(
            name="Fe"), delta=1e-6)
        self.assertAlmostEquals(0.27648, entry.get_element_fraction(
            name="Co"), delta=1e-6)
        self.assertAlmostEquals(0.18432, entry.get_element_fraction(
            name="B"), delta=1e-6)
        self.assertAlmostEquals(0.04608, entry.get_element_fraction(
            name="Si"), delta=1e-6)
        self.assertAlmostEquals(0.0384, entry.get_element_fraction(
            name="Nb"), delta=1e-6)
        self.assertAlmostEquals(0.04, entry.get_element_fraction(
            name="Cr"), delta=1e-6)

    def test_set_composition(self):
        # One element.
        elem = [0]
        frac = [1]
        entry = CompositionEntry(element_ids=elem, fractions=frac)
        self.assertEquals(1, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0, entry.get_element_fraction(name="H"),
                                delta=1e-6)

        # One element with duplicates.
        elem = [0, 0]
        frac = [0.5, 0.5]
        entry = CompositionEntry(element_ids=elem, fractions=frac)
        self.assertEquals(1, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0, entry.get_element_fraction(name="H"),
                                delta=1e-6)

        # One element with zero.
        elem = [0, 1]
        frac = [1, 0]
        entry = CompositionEntry(element_ids=elem, fractions=frac)
        self.assertEquals(1, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0, entry.get_element_fraction(name="H"),
                                delta=1e-6)

        # Two elements.
        elem = [16, 10]
        frac = [1, 1]
        entry = CompositionEntry(element_ids=elem, fractions=frac)
        self.assertEquals(2, len(entry.get_element_ids()))
        self.assertAlmostEquals(0.5, entry.get_element_fraction(name="Na"),
                                delta=1e-6)
        self.assertAlmostEquals(0.5, entry.get_element_fraction(name="Cl"),
                                delta=1e-6)
        np.assert_array_equal([10, 16], entry.get_element_ids())
        np.assert_array_almost_equal([0.5, 0.5], entry.get_element_fractions())
        self.assertAlmostEquals(2, entry.number_in_cell, delta=1e-6)

        # Two elements with duplicates.
        elem = [11, 16, 16]
        frac = [1, 1, 1]
        entry = CompositionEntry(element_ids=elem, fractions=frac)
        self.assertEquals(2, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 3, entry.get_element_fraction(name="Mg"),
                                delta=1e-6)
        self.assertAlmostEquals(2.0 / 3, entry.get_element_fraction(name="Cl"),
                                delta=1e-6)
        np.assert_array_equal([11, 16], entry.get_element_ids())
        np.assert_array_almost_equal([1.0 / 3, 2.0 / 3],
                                     entry.get_element_fractions())
        self.assertAlmostEquals(3, entry.number_in_cell, delta=1e-6)

        # Two elements with zero.
        elem = [11, 16, 16]
        frac = [1, 2, 0]
        entry = CompositionEntry(element_ids=elem, fractions=frac)
        self.assertEquals(2, len(entry.get_element_ids()))
        self.assertAlmostEquals(1.0 / 3, entry.get_element_fraction(name="Mg"),
                                delta=1e-6)
        self.assertAlmostEquals(2.0 / 3, entry.get_element_fraction(name="Cl"),
                                delta=1e-6)
        np.assert_array_equal([11, 16], entry.get_element_ids())
        np.assert_array_almost_equal([1.0 / 3, 2.0 / 3],
                                     entry.get_element_fractions())
        self.assertAlmostEquals(3, entry.number_in_cell, delta=1e-6)

    def test_sort_and_normalize(self):
        # Make an example composition.
        elem = [1, 2, 3, 4, 5]
        frac = [1.0, 2.0, 3.0, 4.0, 5.0]

        # Make first composition.
        entry = CompositionEntry(element_ids=elem, fractions=frac)
        entry_elems = entry.get_element_ids()
        entry_fracs = entry.get_element_fractions()
        for i in range(5):
            self.assertAlmostEquals(entry_fracs[i], entry_elems[i] / 15.0,
                                    delta=1e-6)

        # Iterate through all permutations.
        for perm in permutations([0, 1, 2, 3, 4]):
            # Make a new version of elem and frac.
            new_elem = list(elem)
            new_frac = list(frac)
            for i in range(len(new_elem)):
                new_elem[i] = elem[perm[i]]
                new_frac[i] = frac[perm[i]]

            # Make sure it parses the same.
            new_entry = CompositionEntry(element_ids=elem, fractions=frac)
            self.assertEquals(new_entry, entry)
            self.assertEquals(0, new_entry.__cmp__(entry))
            np.assert_array_equal(entry_elems, new_entry.get_element_ids())
            np.assert_array_almost_equal(entry_fracs,
                                         new_entry.get_element_fractions())

    def test_compare(self):
        this_file_path = os.path.dirname(__file__)
        abs_path = os.path.join(this_file_path, "../../test-files/")
        entries = CompositionEntry.import_composition_list(
            abs_path+"small_set_comp.txt")
        for e1 in range(len(entries)):
            self.assertEquals(0, entries[e1].__cmp__(entries[e1]))
            for e2 in range(e1 + 1, len(entries)):
                self.assertEquals(entries[e1].__cmp__(entries[e2]),
                                  -1 * entries[e2].__cmp__(entries[e1]))
                if entries[e1].__cmp__(entries[e2]) == 0:
                    self.assertEquals(entries[e1].__hash__(), entries[
                        e2].__hash__())
                    self.assertTrue(entries[e1].__eq__(entries[e2]))


# coding=utf-8
import math
from .PairDistanceAnalysis import PairDistanceAnalysis
import numpy as np

class APRDFAnalysis:
    """Class to compute the Atomic Property Weighted Radial Distribution
    Function (AP-RDF).

    Follows the work by Fernandez et al.[1]. Here, we use a scaling factor
    equal to: 1 / #atoms.

    Attributes
    ----------
    structure : Cell
        Link to structure being evaluated.
    cut_off_distance : float
        Cutoff distance used when computing radial distribution function.
    n_windows : int
        Number of points of the RDF to sample.
    B : float
        Smoothing factor in PRDF expression.
    accuracy_factor : float
        Accuracy factor. Determines the number of atoms farther than the
        cutoff radius that are included in the computation of the PRDF. This
        is necessary due to the Gaussian cutoff term in the RDF.

    References
    ----------
    .. [1] M. Fernandez, N. R. Trefiak, and T. K. Woo, "Atomic Property
    Weighted Radial Distribution Functions Descriptors of Metal--Organic
    Frameworks for the Prediction of Gas Uptake Capacity," The Journal of
    Physical Chemistry C, vol. 117, no. 27, pp. 14095--14105, Jul. 2013.

    """
    def __init__(self):
        """Function to create instance and initialize fields.

        Parameters
        ----------
        structure : Cell
            Link to structure being evaluated.
        cut_off_distance : float
            Cutoff distance used when computing radial distribution function.
        n_windows : int
            Number of points of the RDF to sample.
        B : float
            Smoothing factor in PRDF expression.
        accuracy_factor : float
            Accuracy factor. Determines the number of atoms farther than the
            cutoff radius that are included in the computation of the PRDF. This
            is necessary due to the Gaussian cutoff term in the RDF.
        """

        # Link to structure being evaluated.
        self.structure = None

        # Cutoff distance used when computing radial distribution function.
        self.cut_off_distance = 10

        # Number of points of the RDF to sample.
        self.n_windows = 10

        # Smoothing factor in PRDF expression.
        self.B = 2

        # Accuracy factor. Determines the number of atoms farther than the
        # cutoff radius that are included in the computation of the PRDF.
        # This is necessary due to the Gaussian cutoff term in the RDF.
        self.accuracy_factor = 1e-3

    def precompute(self):
        """Function to perform any kind of computations that should be
        performed only once.

        Determine the maximum distance at which atoms contribute
        to the PRDF. This is the maximum distance (r) at which exp(-B * (r -
        R)^2) > AccuracyFactor.

        """
        max_pair_distance = math.sqrt(-1 * math.log(self.accuracy_factor) /
                                      self.B) + self.cut_off_distance

        self.distance_computer = PairDistanceAnalysis()
        self.distance_computer.set_cutoff_distance(max_pair_distance)
        self.distance_computer.analyze_structure(self.structure)

    def analyze_structure(self, s):
        """Function to analyze a specific structure.

        Once this completes, it is possible to retrieve results out of this
        object.

        Parameters
        ----------
        s : Cell
            Structure to be analyzed.

        """
        self.structure = s
        self.precompute()

    def recompute(self):
        """Function to recompute structural information.
        """
        self.precompute()

    def set_smoothing_factor(self, b):
        """Function to set smoothing factor used when computing PRDF.

        Parameters
        ----------
        b : float
            Smoothing factor.

        Raises
        ------
        ValueError
            If B is negative.

        """
        if b <= 0:
            raise ValueError("B must be positive!! Supplied: "+str(b))
        self.B = b

    def set_cut_off_distance(self, d):
        """Function to set cut off distance used when computing PRDF.

        Parameters
        ----------
        d : float
            Cut off distance.

        Raises
        ------
        ValueError
            If d is negative.

        Exception
            If something goes wrong in computing cell.

        """
        if d <= 0:
            raise ValueError("Cut off distance must be positive!! Supplied: "
                             ""+str(d))

        # Set the cutoff distance.
        self.cut_off_distance = d

        # Redo precomputation, if needed.
        if self.structure is not None:
            try:
                self.precompute()
            except Exception:
                raise Exception("Oops, something went wrong. Check "
                                "stacktrace.")

    def set_n_windows(self, n_w):
        """Function to set the number of points at which to evaluate AP-RDF.

        Parameters
        ----------
        n_w : int
            Desired number of windows.

        """
        self.n_windows = n_w

    def compute_APRDF(self, properties):
        """Function to compute the AP-RDF of this structure.

        Parameters
        ----------
        properties : array-like
            Properties of each atom type.

        Returns
        -------
        output : array-like
            AP-RDF at R at n_windows steps between cut off distance /
            n_windows to cut off distance, inclusive.

        """

        # Make sure the number of properties is correct.
        if len(properties) != self.structure.n_types():
            raise ValueError("Incorrect number of properties. Supplied: {}. "
                             "Required: {}".format(len(properties),
                                                   self.structure.n_types()))

        # Get the evaluation distances.
        eval_R = self.get_evaluation_distances()

        # Initialize output.
        ap_rdf = np.zeros(len(eval_R), dtype=float)

        n_atoms = self.structure.n_atoms()
        # Loop through each pair of atoms.
        for i in range(n_atoms):
            for j in range(i, n_atoms):
                # All images of j within cutoff radius of i.
                images = self.distance_computer.find_all_images(i, j)

                # If i != j, the contributions get added twice.
                times_added = 1 if i == j else 2

                # Get types of i and j.
                i_type = self.structure.get_atom(i).get_type()
                j_type = self.structure.get_atom(j).get_type()

                # Add contributes from each pair.
                for image in images:
                    # Evaluate contribution at each distance.
                    for r in range(self.n_windows):
                        ap_rdf[r] += times_added * properties[i_type] * \
                                     properties[j_type] * math.exp(-1 *
                                    self.B * (image[1] - eval_R[r]) ** 2)

        # Scale output by number of atoms.
        ap_rdf *= 1.0 / n_atoms
        return ap_rdf

    def get_evaluation_distances(self):
        """Function to get the distances at which the PRDF should be analyzed.

        Returns
        -------
        output : list
            List of distances.

        """
        step_size = float(self.cut_off_distance) / self.n_windows
        return [(i + 1) * step_size for i in range(self.n_windows)]


import types
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry

class StoichiometricAttributeGenerator:
    """Class to set up and generate descriptors based on the stoichiometry of a
    given material.

    Includes features that are only based on fractions of elements, but not
    what those elements actually are.

    Attributes
    ----------
    p_norms : list
        Exponents to be used in computing various norms.

    """

    # List of p norms to compute.
    def __init__(self, use_default_norms=True):
        """Function to create instance and initialize fields.

        Parameters
        ----------
        use_default_norms : bool
            Whether to use the default norms or not.

        """

        self.p_norms = []
        if use_default_norms:
            self.add_p_norms([2, 3, 5, 7, 10])

    def clear_p_norms(self):
        """Function to clear out the list of p norms to be computed.

        """

        del self.p_norms[:]

    def add_p_norm(self, norm):
        """Function to add a p norm to be computed.

        Parameters
        ----------
        norm : int
            Desired norm.

        Raises
        ------
        ValueError
            If norm is 1.

        """

        if (norm == 0):
            return
        elif (norm == 1):
            raise ValueError("L1 norm is always 1. Useless as attribute.")
        self.p_norms.append(norm)

    def add_p_norms(self, norms):
        """Function to add a list of p norms to be computed.

        Parameters
        ----------
        norm : array-like
            Desired norms. A list of int values.

        """

        for norm in norms:
            self.add_p_norm(norm)

    def generate_features(self, entries):
        """Function to generate the stoichiometric features.

        Computes the norms based on elemental fractions.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.

        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Issue warning if no p norms are added.
        if (not self.p_norms):
            print "Warning: only L0 norm is computed."

        # Add in feature names.
        feat_headers.append("NComp")
        for p in self.p_norms:
            feat_headers.append("Comp_L"+str(p)+"Norm")

        # Compute features.
        for entry in entries:
            fracs = entry.get_element_fractions()
            tmp_list = []
            n_comp = 0
            for f in fracs:
                if (f > 0):
                    n_comp += 1
            # Number of components.
            tmp_list.append(n_comp)

            # Lp norms.
            for p in self.p_norms:
                tmp = 0.0
                for f in fracs:
                    tmp += f**p
                tmp_list.append(tmp**(1.0/p))
            feat_values.append(tmp_list)

        # features as a pandas data frame.
        features = pd.DataFrame(feat_values, columns=feat_headers)
        return features


import numpy as np

class EqualSumCombinations:
    """Class to generate all combinations of non-negative integers that have
    equal sum.

    """
    def __init__(self, sum_, size):
        """Constructor to initialize the variables.

        Parameters
        ----------
        sum_ : int
            Desired sum (must be greater than 0).
        size : int
            Number of integers (must be greater than 1).

        Raises
        ------
        ValueError
            If sum is less than 2.
        """
        if sum_ <= 0:
            raise ValueError("Sum must be positive.")
        if size < 2:
            raise ValueError("Size must be greater than 1.")

        self.dp = np.zeros((sum_+1, size+1), dtype=object)
        self.combs = self.get_combinations(sum_, size)

    def get_combinations(self, sum, n):
        """A recursive function to generate the list of all non-negative
        integer combinations of a given size that have a given sum.

        Parameters
        ----------
        sum : int
            Desired sum.
        n : int
            Desired size.

        Returns
        -------
        type : array-like
            A list containing the lists of combinations.

        """

        # Check if the list of lists has already been generated. If yes,
        # return the value.
        if self.dp[sum][n]:
            return self.dp[sum][n]

        # Initialize list and check special cases first.
        tmp_list = []
        if n == 1:
            tmp_list = [[sum]]
        elif sum == 0:
            tmp_list = [[0] * n]
        else:
            for i in xrange(sum, -1, -1):
                for l in self.get_combinations(sum - i, n - 1):
                    tmp_list.append([i]+l)

        # Store value into the numpy array before returning it.
        self.dp[sum][n] = tmp_list
        return tmp_list


import math
import unittest
from numpy.linalg import norm
import numpy as np
import numpy.testing as np_tst
from ....vassal.data.Atom import Atom
from ....vassal.data.Cell import Cell

class testCell(unittest.TestCase):
    cell = None

    # Create one instance per test.
    def setUp(self):
        self.cell = Cell()

    # Destroy instance as soon as test is over.
    def tearDown(self):
        self.cell = None

    def test_set_basis(self):
        # Test using angles and lattice parameters as input.
        self.cell.set_basis(lengths=[5.643, 6.621,4.885], angles=[91.83,
                            93.58, 107.69])
        self.assertAlmostEquals(173.30, self.cell.volume(), delta=1e-2)
        np_tst.assert_array_almost_equal([5.643, 6.621,4.885],
                                         self.cell.get_lattice_parameters())
        np_tst.assert_array_almost_equal([91.83, 93.58, 107.69],
                    self.cell.get_lattice_angles_radians(radians=False))

        # Simple test with a primitive cell.
        basis = np.zeros((3, 3))
        basis[0] = np.array([0, 2.986, 2.986])
        basis[1] = np.array([2.986, 0, 2.986])
        basis[2] = np.array([2.986, 2.986, 0])

        self.cell.set_basis(basis=basis)
        self.assertAlmostEquals(13.312*4, self.cell.volume(), delta=1e-3)
        np_tst.assert_array_almost_equal([4.223, 4.223, 4.223],
                                         self.cell.get_lattice_parameters(),
                                         decimal=3)
        np_tst.assert_array_almost_equal([60, 60, 60],
                    self.cell.get_lattice_angles_radians(radians=False))

    def test_aligned_basis(self):
        # Simple test with a primitive cell.
        basis = np.zeros((3, 3))
        basis[0] = np.array([0, 2.986, 2.986])
        basis[1] = np.array([2.986, 0, 2.986])
        basis[2] = np.array([2.986, 2.986, 0])

        self.cell.set_basis(basis=basis)

        # Compute the aligned basis.
        aligned_basis = self.cell.get_aligned_basis()
        self.assertAlmostEquals(0, aligned_basis[1][0], delta=1e-6)
        self.assertAlmostEquals(0, aligned_basis[2][0], delta=1e-6)
        self.assertAlmostEquals(0, aligned_basis[2][1], delta=1e-6)

    def test_clone(self):
        self.cell.add_atom(Atom([0, 0, 0], 0))
        self.cell.set_type_name(0, "A")

        # Test adding atoms.
        clone = self.cell.__copy__()
        self.assertEquals(clone, self.cell)
        clone.add_atom(Atom([0, 0.5, 0], 0))
        self.assertFalse(clone.__eq__(self.cell))

        # Test changing atom.
        clone = self.cell.__copy__()
        clone.get_atom(0).set_type(1)
        self.assertFalse(clone.__eq__(self.cell))

        # Test changing basis.
        clone = self.cell.__copy__()
        clone.set_basis(lengths=[2, 1, 1], angles=[90, 90, 90])
        self.assertFalse(clone.__eq__(self.cell))

    def test_lattice_vectors(self):
       self.cell.set_basis(lengths=[1, 2, 3], angles=[80, 90, 70])
       l_vec = self.cell.get_lattice_vectors()
       np_tst.assert_array_almost_equal([[1.0, 0.0, 0.0], [0.684, 1.879,
                0.0], [0.0, 0.554, 2.948]], l_vec, decimal=3)

       # FCC primitive cell.
       self.cell.set_basis(lengths=[0.70710678118655, 0.70710678118655,
                                    0.70710678118655], angles=[60, 60, 60])
       self.assertAlmostEquals(0.25, self.cell.volume(), delta=1e-6)
       l_vec = self.cell.get_lattice_vectors()
       self.assertAlmostEquals(0.70710678118655, norm(l_vec[0]),
                               delta=1e-2)

    def test_fractional_to_cartesian(self):
        self.cell.set_basis(lengths=[1, 2, 3], angles=[80, 90, 70])
        np_tst.assert_array_almost_equal([0.2368, 0.5421, 0.8844],
                self.cell.convert_fractional_to_cartesian([0.1, 0.2, 0.3]),
                                         decimal=3)

    def test_cartesian_to_fractional(self):
        self.cell.set_basis(lengths=[1, 2, 3], angles=[80, 90, 70])
        np_tst.assert_array_almost_equal([0.1, 0.2, 0.3],
            self.cell.convert_cartesian_to_fractional([0.2368, 0.5421, 0.8844]),
                                         decimal=3)

    def test_supercell_translation(self):
        self.cell.set_basis(lengths=[0.70710678118655, 0.70710678118655,
                                     0.70710678118655], angles=[60, 60, 60])
        self.assertAlmostEquals(0.25, self.cell.volume(), delta=1e-6)
        l_vec = self.cell.get_lattice_vectors()

        # Check a few.
        pos = self.cell.get_periodic_image([0, 0, 0], 1, 0, 0)
        np_tst.assert_array_almost_equal([0.70710678000000, 0, 0], pos,
                                         decimal=3)
        pos = self.cell.get_periodic_image([0, 0, 0], 1, 1, 0)
        np_tst.assert_array_almost_equal([1.06066017000000, 0.61237243466821,
                                0], pos, decimal=3)
        pos = self.cell.get_periodic_image([0, 0, 0], 1, 1, 1)
        np_tst.assert_array_almost_equal([1.41421356000000, 0.81649657955762,
                                0.57735026918963], pos, decimal=3)

    def test_equals(self):
        # Make other cell
        other = Cell()

        # First check.
        self.assertTrue(self.cell.__eq__(other))

        # Adjust basis.
        self.cell.set_basis(lengths=[1, 2, 3], angles=[70, 80, 90])
        self.assertFalse(self.cell.__eq__(other))
        other.set_basis(lengths=[1, 2, 3], angles=[70, 80, 90])
        self.assertTrue(self.cell.__eq__(other))

        # Add an atom to 0,0,0
        self.cell.add_atom(Atom([0, 0, 0], 0))
        self.assertFalse(self.cell.__eq__(other))
        other.add_atom(Atom([0, 0, 0], 0))
        self.assertTrue(self.cell.__eq__(other))

        # Changing names.
        self.cell.set_type_name(0, "Al")
        self.assertFalse(self.cell.__eq__(other))
        other.set_type_name(0, "Al")
        self.assertTrue(self.cell.__eq__(other))

        # Adding more atoms of different type.
        self.cell.add_atom(Atom([0.5, 0.5, 0], 1))
        other.add_atom(Atom([0.5, 0.5, 0], 0))
        self.assertFalse(self.cell.__eq__(other))
        other.get_atom(1).set_type(1)
        self.assertTrue(self.cell.__eq__(other))

        # Adding atoms with different positions.
        self.cell.add_atom(Atom([0.5, 0, 0.5], 1))
        other.add_atom(Atom([0, 0.5, 0.5], 1))
        self.assertFalse(self.cell.__eq__(other))

        # Adding atoms out of sequence.
        other.add_atom(Atom([0.5, 0, 0.5], 1))
        self.cell.add_atom(Atom([0, 0.5, 0.5], 1))
        self.assertTrue(self.cell.__eq__(other))

    def test_minimum_distance(self):
        # Simple case: orthogonal axes.

        # Origin.
        self.cell.add_atom(Atom([0, 0, 0], 1))
        # C face center.
        self.cell.add_atom(Atom([0.5, 0.5, 0], 1))

        dist = self.cell.get_minimum_distance(point1=[0, 0, 0], point2=[0.5,
                                                            0.5, 0])
        self.assertAlmostEquals(math.sqrt(0.5), dist, delta=1e-6)
        dist = self.cell.get_minimum_distance(point1=[0, 0, 0], point2=[2.5,
                                                            0.5, -10])
        self.assertAlmostEquals(math.sqrt(0.5), dist, delta=1e-6)

        # Difficult case: Non-conventional unit cell.
        basis = self.cell.get_basis()
        basis[1][0] = 108
        self.cell.set_basis(basis=basis)
        dist = self.cell.get_minimum_distance(point1=[0, 0, 0], point2=[0.5,
                                                            0.5, 0])
        self.assertAlmostEquals(math.sqrt(0.5), dist, delta=1e-6)
        dist = self.cell.get_minimum_distance(point1=[0, 0, 0], point2=[5.5,
                                                            0.5, 0])
        self.assertAlmostEquals(math.sqrt(0.5), dist, delta=1e-6)
        dist = self.cell.get_minimum_distance(point1=[0, 0, 0], point2=[5.5,
                                                            -10.5, 0])
        self.assertAlmostEquals(math.sqrt(0.5), dist, delta=1e-6)

    def test_get_closest_image_simple(self):
        # Simple case: orthogonal axes.

        # Origin.
        self.cell.add_atom(Atom([0, 0, 0], 1))
        # C face center.
        self.cell.add_atom(Atom([0.75, 0.75, 0.75], 1))
        image = self.cell.get_minimum_distance(center=0, neighbor=1)
        np_tst.assert_array_almost_equal([-0.25, -0.25, -0.25],
                                         image.get_position(), decimal=6)
        np_tst.assert_array_equal([-1, -1, -1], image.get_supercell())

    def test_get_closest_image_difficult(self):
        # Difficult case: Non-conventional unit cell.
        # Origin.
        self.cell.add_atom(Atom([0, 0, 0], 1))
        # Body face center.
        self.cell.add_atom(Atom([0.5, 0.5, 0.5], 1))
        basis = self.cell.get_basis()
        basis[1][0] = 108
        self.cell.set_basis(basis=basis)
        image = self.cell.get_minimum_distance(center=0, neighbor=1)
        np_tst.assert_array_almost_equal([-0.5, -0.5, 0.5],
                                         image.get_position(), decimal=6)
        np_tst.assert_array_equal([-1, 53, 0], image.get_supercell())

    def test_replacement(self):
        # Make the original cell B2-CuZr
        self.cell.add_atom(Atom([0, 0, 0], 0))
        self.cell.add_atom(Atom([0.5, 0.5, 0.5], 1))
        self.cell.set_type_name(0, "Cu")
        self.cell.set_type_name(1, "Zr")

        # Replace Cu with Ni.
        to_change = {"Cu":"Ni"}
        self.cell.replace_type_names(to_change)
        self.assertEquals("Ni", self.cell.get_type_name(0))
        self.assertEquals("Zr", self.cell.get_type_name(1))

        # Replace Ni with Cu and Zr with Ti.
        to_change = {"Ni": "Cu", "Zr":"Ti"}
        self.cell.replace_type_names(to_change)
        self.assertEquals("Cu", self.cell.get_type_name(0))
        self.assertEquals("Ti", self.cell.get_type_name(1))

        # Exchange Cu and Ti.
        to_change = {"Ti": "Cu", "Cu": "Ti"}
        self.cell.replace_type_names(to_change)
        self.assertEquals("Ti", self.cell.get_type_name(0))
        self.assertEquals("Cu", self.cell.get_type_name(1))

        # Make everything Cu.
        to_change = {"Ti": "Cu"}
        self.cell.replace_type_names(to_change)
        self.assertEquals("Cu", self.cell.get_type_name(0))
        self.assertEquals("Cu", self.cell.get_type_name(1))

        # Make everything W.
        to_change = {"Cu":"W"}
        self.cell.replace_type_names(to_change)
        self.assertEquals("W", self.cell.get_type_name(0))
        self.assertEquals("W", self.cell.get_type_name(1))

        # Merge types.
        self.cell.merge_like_types()
        self.assertEquals(1, self.cell.n_types())


import unittest
from ....data.materials.CrystalStructureEntry import CrystalStructureEntry
from ....vassal.data.Atom import Atom
from ....vassal.data.Cell import Cell

class testAtomicStructureEntry(unittest.TestCase):
    def test_replacement(self):
        # Make B2-CuZr
        cell = Cell()
        cell.add_atom(Atom([0, 0, 0], 0))
        cell.add_atom(Atom([0.5, 0.5, 0.5], 1))
        cell.set_type_name(0, "Cu")
        cell.set_type_name(1, "Zr")
        CuZr = CrystalStructureEntry(cell, "CuZr", None)

        # Run Voronoi tessellation.
        CuZr.compute_voronoi_tessellation()

        # Make B2-NiZr
        changes = {"Cu":"Ni"}
        NiZr = CuZr.replace_elements(changes)

        # Make sure the tessellation object did not change.
        self.assertTrue(CuZr.compute_voronoi_tessellation() is
                        NiZr.compute_voronoi_tessellation())

        # Make sure the two are still unchanged.
        self.assertAlmostEquals(0.5, CuZr.get_element_fraction(name="Cu"),
                                delta=1e-6)
        self.assertAlmostEquals(0.0, CuZr.get_element_fraction(name="Ni"),
                                delta=1e-6)
        self.assertAlmostEquals(0.0, NiZr.get_element_fraction(name="Cu"),
                                delta=1e-6)
        self.assertAlmostEquals(0.5, NiZr.get_element_fraction(name="Ni"),
                                delta=1e-6)

        # Now, change the structure such that it has fewer types.
        changes["Ni"] = "Zr"

        bccZr = NiZr.replace_elements(changes)

        # Make sure the structure only has one type.
        self.assertAlmostEquals(1.0, bccZr.get_element_fraction(name="Zr"),
                                delta=1e-6)
        self.assertEquals(1, bccZr.get_structure().n_types())

        self.assertFalse(NiZr.compute_voronoi_tessellation() is
                         bccZr.compute_voronoi_tessellation())

    def test_to_string(self):
        # Make B2-CuZr
        cell = Cell()
        cell.add_atom(Atom([0, 0, 0], 0))
        cell.add_atom(Atom([0.5, 0.5, 0.5], 1))
        cell.set_type_name(0, "Cu")
        cell.set_type_name(1, "Zr")
        CuZr = CrystalStructureEntry(cell, "B2", None)

        name = CuZr.__str__()
        print name
        self.assertTrue("CuZr" in name)
        self.assertTrue("B2" in name)


# coding=utf-8
from itertools import izip
import numpy as np
import pandas as pd
from scipy.linalg import lu
from scipy.optimize import linprog
from .LookUpData import LookUpData
from ..CompositionEntry import CompositionEntry

class GCLPCalculator:
    """Class that uses Grand-Canonical Linear Programming (GCLP) method to
    compute phase equilibria.

    How to use this class:
    1. Load in phase equilibria data with `self.add_phase`, `self.set_mu` or
    `self.add_phases`.
    2. Compute equilibrium with `self.run_GCLP`.

    Attributes
    ----------
    lp_element_names : array-like
        List of element names.
    phases : dict
        Dictionary containing CompositionEntry as the key and energy (float)
        as the value. Phases to consider for equilibria and their energy.
        Only contains the lowest-energy phase at each entry.

    References
    ----------
    .. [1] A. R. Akbarzadeh, V. Ozoliņš, and C. Wolverton, “First-Principles
    Determination of Multicomponent Hydride Phase Diagrams: Application to
    the Li-Mg-N-H System,” Advanced Materials, vol. 19, no. 20,
    pp. 3233--3239, Sep. 2007.

    """
    lp_element_names = LookUpData.element_names

    def __init__(self):
        """Function to create instance and initialize fields.

        Initialize a GCLP calculator. Sets the chemical potential of each
        element to be 0.
        """
        self.phases = {}
        # Add for each element.
        for elem in self.lp_element_names:
            entry = CompositionEntry(element_names=[elem], fractions=[1.0])
            self.phases[entry] = 0.0

    def set_mu(self, elem, mu):
        """Function to set the chemical potential of an element.

        Parameters
        ----------
        elem : str
            Abbreviation of element.

        mu : float
            Desired chemical potential.

        """
        entry = CompositionEntry(composition=elem)
        if len(entry.get_element_ids()) != 1:
            raise ValueError("Not an element "+elem)
        self.phases[entry] = mu

    def add_phases(self, entries, energies):
        """Function to set many phase energies.

        Parameters
        ----------
        entries : array-like
            A list of CompositionEntry's.

        energies : array-like
            A list of energy values as floats.

        """
        if isinstance(energies,pd.DataFrame):
            if energies.shape[1] == 1:
                energies = energies.values
            else:
                msg = "a dataframe with one column is required"
                raise Exception(msg)
        if isinstance(entries,pd.DataFrame):
            if entries.shape[1] == 1:
                entries = entries.values
            else:
                msg = "a dataframe with one column is required"
                raise Exception(msg)

        for entry,energy in izip(entries, energies):
            # if has measurement
            self.add_phase(entry, energy)

    def add_phase(self, entry, energy):
        """Function to set one phase.

        Parameters
        ----------
        entry : CompositionEntry
            Composition of phase.

        energy : float
            Energy of phase.


        """
        if entry not in self.phases:
            # Add if there is no current entry at this composition.
            self.phases[entry] = float(energy)
        elif self.phases[entry] > energy:
            # If there is a phase, update only if new energy is lower than
            # current.
            self.phases[entry] = float(energy)

    def get_num_phases(self):
        """
        Function to get the number of phases being considered for GCLP.

        Returns
        -------
        output : int
            The number of phases being considered.

        """
        return len(self.phases)

    def run_GCLP(self, composition):
        """Function to compute the ground state phase equilibria for a
        certain composition.

        Parameters
        ----------
        composition : CompositionEntry
            Composition to be considered.

        Returns
        -------
        ground_state_energy : float
            Ground state energy.
        equilibrium : dict
            Dictionary containing the phase composition (CompositionEntry) as
            key and the fractions (float) as values.

        Raises
        ------
        TypeError
            If composition is not CompositionEntry.

        """
        if not isinstance(composition, CompositionEntry):
            raise TypeError("Composition should be of type CompositionEntry!")

        cur_elements = composition.get_element_ids()
        cur_fractions = composition.get_element_fractions()

        # List of composition entries.
        components = []

        # List of energies.
        energies = []

        # Get the current possible phases (i.e., those that contain
        # exclusively the elements in the current compound.
        for entry in self.phases:
            this_elements = entry.get_element_ids()
            # Check whether this entry is in the target phase diagram.
            if set(this_elements) <= set(cur_elements):
                components.append(entry)
                energies.append(self.phases[entry])

        # Set up constraints.
        # Type #1: Mass conservation.
        l_components = len(components)
        l_composition = len(cur_elements)
        a_eq = np.ones((l_composition + 1, l_components))
        b_eq = np.ones(l_composition + 1)
        for i in range(l_composition):
            b_eq[i] = cur_fractions[i]
            for j in range(l_components):
                a_eq[i][j] = components[j].get_element_fraction(
                    id=cur_elements[i])

        # Type #2: Normalization.
        # Taken care of when we initialized a_eq and b_eq to ones.

        # Perform LU decomposition to check if there are any linearly
        # dependent rows in the matrix a_eq. For some reason, linprog can't
        # handle linearly dependent matrices.
        _, u = lu(a_eq, permute_l=True)

        mask = np.all(abs(u) < 1e-14, axis=1)
        indices = [i for i in range(len(mask)) if mask[i]]
        if indices:
            a_eq = np.delete(a_eq, indices, axis=0)
            b_eq = np.delete(b_eq, indices)

        c = np.array(energies)

        # Call LP solver and store result.
        res = linprog(c=c, A_eq=a_eq, b_eq=b_eq)
        equilibrium = {}
        equilibrium_fractions = res.x
        for i in range(l_components):
            if equilibrium_fractions[i] > 1e-6:
                equilibrium[components[i]] = equilibrium_fractions[i]

        # Add zero to avoid returning -0.0 values.
        ground_state_energy = res.fun + 0

        return ground_state_energy, equilibrium